"""ThreatConnect Artifact"""

from .api_endpoints import ApiEndpoints
from .common_case_management import CommonCaseManagement
from .common_case_management_collection import CommonCaseManagementCollection
from .filter import Filter
from .tql import TQL


class Artifacts(CommonCaseManagementCollection):
    """Artifacts Class for Case Management Collection

    .. code-block:: python
        :linenos:
        :lineno-start: 1

        # Example of params input
        {
            'result_limit': 100,  # How many results are retrieved.
            'result_start': 10,  # Starting point on retrieved results.
            'fields': ['caseId', 'summary']  # Additional fields returned on the results
        }

    Args:
        tcex (TcEx): An instantiated instance of TcEx object.
        initial_response (dict, optional): Initial data in
            Case Object for Artifact. Defaults to None.
        tql_filters (list, optional): List of TQL filters. Defaults to None.
        params(dict, optional): Dict of the params to be sent while
            retrieving the Artifacts objects.
    """

    def __init__(self, tcex, initial_response=None, tql_filters=None, params=None):
        """Initialize Class properties"""
        super().__init__(
            tcex,
            ApiEndpoints.ARTIFACTS,
            initial_response=initial_response,
            tql_filters=tql_filters,
            params=params,
        )

        if initial_response:
            for item in initial_response.get('data', []):
                self.added_items.append(Artifact(tcex, **item))

    def __iter__(self):
        """Object iterator"""
        return self.iterate(initial_response=self.initial_response)

    def add_artifact(self, artifact):
        """Add an Artifact.

        Args:
            artifact (Artifact): The Artifact Object to add.
        """
        self.added_items.append(artifact)

    def entity_map(self, entity):
        """Map a dict to a Artifact.

        Args:
            entity (dict): The Artifact data.

        Returns:
            CaseManagement.Artifact: An Artifact Object
        """
        return Artifact(self.tcex, **entity)

    @property
    def filter(self):
        """Return instance of FilterArtifact Object."""
        return FilterArtifacts(ApiEndpoints.ARTIFACTS, self.tcex, self.tql)


class Artifact(CommonCaseManagement):
    """Artifact object for Case Management.

    Args:
        tcex (TcEx): An instantiated instance of TcEx object.
        analytics_priority (str, kwargs): [Read-Only] The **Analytics Priority** for the Artifact.
        analytics_priority_level (int, kwargs): [Read-Only] The **Analytics Priority Level** for
            the Artifact.
        analytics_score (int, kwargs): [Read-Only] The **Analytics Score** for the Artifact.
        analytics_status (int, kwargs): [Read-Only] The **Analytics Status** for the Artifact.
        analytics_type (str, kwargs): [Read-Only] The **Analytics Type** for the Artifact.
        artifact_type (ArtifactType, kwargs): [Read-Only] The **Artifact Type** for the Artifact.
        case_id (int, kwargs): [Required (alt: caseXid)] The **Case Id** for the Artifact.
        case_xid (str, kwargs): [Required (alt: caseId)] The **Case Xid** for the Artifact.
        date_added (str, kwargs): [Read-Only] The **Date Added** for the Artifact.
        file_data (str, kwargs): Base64 encoded file attachment required only for certain artifact
            types
        intel_type (str, kwargs): The **Intel Type** for the Artifact.
        links (Link, kwargs): The **Links** for the Artifact.
        notes (Note, kwargs): a list of Notes corresponding to the Artifact
        parent_case (Case, kwargs): [Read-Only] The **Parent Case** for the Artifact.
        source (str, kwargs): The **Source** for the Artifact.
        summary (str, kwargs): [Required] The **Summary** for the Artifact.
        task (case_management.task.Task, kwargs): [Read-Only] The **Task** for the Artifact.
        task_id (int, kwargs): the ID of the task which the Artifact references
        task_xid (str, kwargs): the XID of the task which the Artifact references
        type (str, kwargs): [Required] The **Type** for the Artifact.
    """

    def __init__(self, tcex, **kwargs):
        """Initialize Class properties"""
        super().__init__(tcex, ApiEndpoints.ARTIFACTS, kwargs)
        self.artifact_filter = [
            {
                'keyword': 'artifactId',
                'operator': TQL.Operator.EQ,
                'value': self.id,
                'type': TQL.Type.INTEGER,
            }
        ]

        self._analytics_priority = kwargs.get('analytics_priority', None)
        self._analytics_priority_level = kwargs.get('analytics_priority_level', None)
        self._analytics_score = kwargs.get('analytics_score', None)
        self._analytics_status = kwargs.get('analytics_status', None)
        self._analytics_type = kwargs.get('analytics_type', None)
        self._artifact_type = kwargs.get('artifact_type', None)
        self._case_id = kwargs.get('case_id', None) or kwargs.get('parent_case', {}).get('id', None)
        self._case_xid = kwargs.get('case_xid', None)
        self._date_added = kwargs.get('date_added', None)
        self._field_name = kwargs.get('field_name', None)
        self._file_data = kwargs.get('file_data', None)
        self._intel_type = kwargs.get('intel_type', None)
        self._links = kwargs.get('links', None)
        self._notes = kwargs.get('notes', None)
        self._parent_case = kwargs.get('parent_case', None)
        self._source = kwargs.get('source', None)
        self._summary = kwargs.get('summary', None)
        self._task = kwargs.get('task', None)
        self._task_id = kwargs.get('task_id', None)
        self._task_xid = kwargs.get('task_xid', None)
        self._type = kwargs.get('type', None)

    def add_note(self, **kwargs):
        """Add a Note to a Case."""
        self.notes.add_note(self.tcex.cm.note(**kwargs))

    @property
    def analytics_priority(self):
        """Return the **Analytics Priority** for the Artifact."""
        return self._analytics_priority

    @property
    def analytics_priority_level(self):
        """Return the **Analytics Priority Level** for the Artifact."""
        return self._analytics_priority_level

    @property
    def analytics_score(self):
        """Return the **Analytics Score** for the Artifact."""
        return self._analytics_score

    @property
    def analytics_status(self):
        """Return the **Analytics Status** for the Artifact."""
        return self._analytics_status

    @property
    def analytics_type(self):
        """Return the **Analytics Type** for the Artifact."""
        return self._analytics_type

    @property
    def artifact_type(self):
        """Return the **Artifact Type** for the Artifact."""
        if self._artifact_type:
            return self.tcex.cm.artifact_type(**self._artifact_type)
        return self._artifact_type

    @property
    def as_entity(self):
        """Return the entity representation of the Artifact."""
        return {'type': 'Artifact', 'value': self.summary, 'id': self.id}

    @property
    def case_id(self):
        """Return the parent **Case ID** for the Artifact."""
        return self._case_id

    @case_id.setter
    def case_id(self, case_id):
        """Set the parent **Case ID** for the Artifact."""
        self._case_id = case_id

    @property
    def case_xid(self):
        """Return the **Case XID** for the Artifact."""
        return self._case_xid

    @case_xid.setter
    def case_xid(self, case_xid):
        """Set the **Case XID** for the Artifact."""
        self._case_xid = case_xid

    @property
    def date_added(self):
        """Return the **Date Added** for the Artifact."""
        return self._date_added

    def entity_mapper(self, entity):
        """Update current object with provided object properties.

        Args:
            entity (dict): An entity dict used to update the Object.
        """
        new_artifact = Artifact(self.tcex, **entity)
        self.__dict__.update(new_artifact.__dict__)

    @property
    def field_name(self):
        """Return the **File Data** for the Artifact."""
        return self._field_name

    @property
    def file_data(self):
        """Return the **File Data** for the Artifact."""
        return self._file_data

    @file_data.setter
    def file_data(self, file_data):
        """Set the **File Data** for the Artifact."""
        self._file_data = file_data

    @property
    def intel_type(self):
        """Return the **Intel Type** for the Artifact."""
        return self._intel_type

    @property
    def links(self):
        """Return the **Links** for the Artifact."""
        # TODO: create an object for this
        return self._links

    @property
    def notes(self):
        """Return the **Notes** for the Artifact."""
        if self._notes is None or isinstance(self._notes, dict):
            notes = self._notes or {}
            self._notes = self.tcex.cm.notes(
                initial_response=notes, tql_filters=self.artifact_filter
            )
        return self._notes

    @notes.setter
    def notes(self, notes):
        """Set the **Notes** for the Artifact."""
        if isinstance(notes, dict):
            self._notes = self.tcex.cm.notes(
                initial_response=notes, tql_filters=self.artifact_filter
            )
        self._notes = notes

    @property
    def parent_case(self):
        """Return the **Parent Case** for the Task."""
        if self._parent_case:
            return self.tcex.cm.case(**self._parent_case)
        return self._parent_case

    @property
    def source(self):
        """Return the **Source** for the Artifact."""
        return self._source

    @source.setter
    def source(self, source):
        """Set the **Source** for the Artifact."""
        self._source = source

    @property
    def summary(self):
        """Return the **Summary** for the Artifact."""
        return self._summary

    @summary.setter
    def summary(self, summary):
        """Set the **Summary** for the Artifact."""
        self._summary = summary

    @property
    def task(self):
        """Return the **Task** for the Note."""
        if self._task:
            return self.tcex.cm.task(**self._task)
        return self._task

    @property
    def task_id(self):
        """Return the **Task ID** for the Artifact."""
        return self._task_id

    @task_id.setter
    def task_id(self, task_id):
        """Set the **Task ID** for the Artifact."""
        self._task_id = task_id

    @property
    def task_xid(self):
        """Return the **Task XID** for the Artifact."""
        return self._task_xid

    @task_xid.setter
    def task_xid(self, task_xid):
        """Set the **Task XID** for the Artifact."""
        self._task_xid = task_xid

    @property
    def type(self):
        """Return the **Type** for the Artifact."""
        return self._type

    @type.setter
    def type(self, artifact_type):
        """Set the **Type** for the Artifact."""
        self._type = artifact_type


class FilterArtifacts(Filter):
    """Filter Object for Artifacts"""

    def case_id(self, operator, case_id):
        """Filter Artifacts based on **caseId** keyword.

        Args:
            operator (enum): The operator enum for the filter.
            case_id (int): The ID of the case associated with this artifact.
        """
        self._tql.add_filter('caseId', operator, case_id, TQL.Type.INTEGER)

    @property
    def has_case(self):
        """Return **FilterCases** for further filtering."""
        from .case import FilterCases

        cases = FilterCases(ApiEndpoints.CASES, self._tcex, TQL())
        self._tql.add_filter('hasCase', TQL.Operator.EQ, cases, TQL.Type.SUB_QUERY)
        return cases

    @property
    def has_note(self):
        """Return **FilterNotes** for further filtering."""
        from .note import FilterNotes  # pylint: disable=cyclic-import

        notes = FilterNotes(ApiEndpoints.NOTES, self._tcex, TQL())
        self._tql.add_filter('hasNote', TQL.Operator.EQ, notes, TQL.Type.SUB_QUERY)
        return notes

    @property
    def has_task(self):
        """Return **FilterTask** for further filtering."""
        from .task import FilterTasks  # pylint: disable=cyclic-import

        tasks = FilterTasks(ApiEndpoints.TASKS, self._tcex, TQL())
        self._tql.add_filter('hasTask', TQL.Operator.EQ, tasks, TQL.Type.SUB_QUERY)
        return tasks

    def id(self, operator, id):  # pylint: disable=redefined-builtin
        """Filter Artifacts based on **id** keyword.

        Args:
            operator (enum): The operator enum for the filter.
            id (int): The ID of the artifact.
        """
        self._tql.add_filter('id', operator, id, TQL.Type.INTEGER)

    def note_id(self, operator, note_id):
        """Filter Artifacts based on **noteId** keyword.

        Args:
            operator (enum): The operator enum for the filter.
            note_id (int): The ID of the note associated with this artifact.
        """
        self._tql.add_filter('noteId', operator, note_id, TQL.Type.INTEGER)

    def source(self, operator, source):
        """Filter Artifacts based on **source** keyword.

        Args:
            operator (enum): The operator enum for the filter.
            source (str): The source of the artifact.
        """
        self._tql.add_filter('source', operator, source, TQL.Type.STRING)

    def summary(self, operator, summary):
        """Filter Artifacts based on **summary** keyword.

        Args:
            operator (enum): The operator enum for the filter.
            summary (str): The summary of the artifact.
        """
        self._tql.add_filter('summary', operator, summary, TQL.Type.STRING)

    def task_id(self, operator, task_id):
        """Filter Artifacts based on **taskId** keyword.

        Args:
            operator (enum): The operator enum for the filter.
            task_id (int): The ID of the task associated with this artifact.
        """
        self._tql.add_filter('taskId', operator, task_id, TQL.Type.INTEGER)

    def type_name(self, operator, type_name):
        """Filter Artifacts based on **typeName** keyword.

        Args:
            operator (enum): The operator enum for the filter.
            type_name (str): The type name of the artifact.
        """
        self._tql.add_filter('typeName', operator, type_name, TQL.Type.STRING)



#!/usr/bin/env python
"""TcEx Framework LayoutJson."""
# standard library
import json
import logging
import os
from collections import OrderedDict


class LayoutJson:
    """Object for layout.json file.

    Args:
        filename (str, optional): The config filename. Defaults to layout.json.
        path (str, optional): The path to the file. Defaults to os.getcwd().
        logger (logging.Logger, optional): A instance of Logger. Defaults to None.
    """

    def __init__(self, filename=None, path=None, logger=None):
        """Initialize class properties."""
        self._filename = filename or 'layout.json'
        self._path = path or os.getcwd()
        self.log = logger or logging.getLogger('layout_json')

        # properties
        self._contents = None

    @staticmethod
    def _to_bool(value):
        """Convert string value to bool."""
        bool_value = False
        if str(value).lower() in ['1', 'true']:
            bool_value = True
        return bool_value

    @property
    def contents(self):
        """Return layout.json contents."""
        if self._contents is None and self.has_layout:
            with open(self.filename) as fh:
                self._contents = json.load(fh, object_pairs_hook=OrderedDict)
        return self._contents

    def create(self, inputs, outputs):
        """Create new layout.json file based on inputs and outputs."""
        lj = OrderedDict()

        # add inputs
        lj['inputs'] = []
        step = OrderedDict()
        step['parameters'] = []
        step['sequence'] = 1
        step['title'] = 'Action'
        lj['inputs'].append(step)
        step = OrderedDict()
        step['parameters'] = []
        step['sequence'] = 2
        step['title'] = 'Connection'
        lj['inputs'].append(step)
        step = OrderedDict()
        step['parameters'] = []
        step['sequence'] = 3
        step['title'] = 'Configure'
        lj['inputs'].append(step)
        step = OrderedDict()
        step['parameters'] = []
        step['sequence'] = 4
        step['title'] = 'Advanced'
        lj['inputs'].append(step)

        # add outputs
        lj['outputs'] = []

        for i in sorted(inputs):
            if i.get('name') == 'tc_action':
                lj['inputs'][0]['parameters'].append({'name': 'tc_action'})
            elif i.get('hidden') is True:
                lj['inputs'][2]['parameters'].append(
                    {'display': "'hidden' != 'hidden'", 'hidden': 'true', 'name': i.get('name')}
                )
            else:
                lj['inputs'][2]['parameters'].append({'display': '', 'name': i.get('name')})

        for o in sorted(outputs):
            lj['outputs'].append({'display': '', 'name': o.get('name')})

        # write layout file to disk
        self.write(lj)

    @property
    def filename(self):
        """Return the fqpn for the layout.json file."""
        return os.path.join(self._path, self._filename)

    @property
    def has_layout(self):
        """Return True if App has layout.json file."""
        if os.path.isfile(self.filename):
            return True
        return False

    @property
    def params_dict(self):
        """Return layout.json params in a flattened dict with name param as key."""
        parameters = {}
        for i in self.inputs:
            for p in i.get('parameters', []):
                parameters.setdefault(p.get('name'), p)
        return parameters

    @property
    def parameters_names(self):
        """Return layout.json params in a flattened dict with name param as key."""
        return self.params_dict.keys()

    @property
    def outputs_dict(self):
        """Return layout.json outputs in a flattened dict with name param as key."""
        outputs = {}
        for o in self.outputs:
            outputs.setdefault(o.get('name'), o)
        return outputs

    def update(self, features=None):
        """Update the layouts.json file."""
        # features from the instal.json
        features = features or []

        # get fresh copy of layout.json contents
        layout_data = self.contents

        # APP-86 - sort output data by name
        self.update_sort_outputs(layout_data)

        # update contents
        self._contents = layout_data

        # write updated content
        self.write(layout_data)

    @staticmethod
    def update_sort_outputs(layout_data):
        """Sort output field by name."""
        # APP-86 - sort output data by name
        layout_data['outputs'] = sorted(layout_data.get('outputs', []), key=lambda i: i['name'])

    def write(self, json_data):
        """Write updated profile file.

        Args:
            json_data (dict): The profile data.
        """
        with open(self.filename, 'w') as fh:
            json.dump(json_data, fh, indent=2, sort_keys=False)
            fh.write('\n')

    #
    # properties
    #

    @property
    def inputs(self):
        """Return property."""
        return self.contents.get('inputs', [])

    @property
    def outputs(self):
        """Return property."""
        return self.contents.get('outputs', [])



"""ThreatConnect Case Management"""
from .artifact import Artifact, Artifacts
from .artifact_type import ArtifactType, ArtifactTypes
from .assignee import Assignee, User, Users
from .case import Case, Cases
from .note import Note, Notes
from .tag import Tag, Tags
from .task import Task, Tasks
from .workflow_event import WorkflowEvent, WorkflowEvents
from .workflow_template import WorkflowTemplate, WorkflowTemplates


class CaseManagement:
    """Case Management Class for TcEx

    Args:
        tcex (TcEx): An instantiated instance of TcEx object.
    """

    def __init__(self, tcex):
        """Initialize Class properties."""
        self.tcex = tcex

    def artifact(self, **kwargs):
        """Return a instance of Artifact object.

        Args:
            tcex (TcEx): An instantiated instance of TcEx object.
            analytics_priority (str, kwargs): [Read-Only] The **Analytics Priority**
                for the Artifact.
            analytics_priority_level (int, kwargs): [Read-Only] The **Analytics Priority Level** for
                the Artifact.
            analytics_score (int, kwargs): [Read-Only] The **Analytics Score** for the Artifact.
            analytics_status (int, kwargs): [Read-Only] The **Analytics Status** for the Artifact.
            analytics_type (str, kwargs): [Read-Only] The **Analytics Type** for the Artifact.
            artifact_type (ArtifactType, kwargs): [Read-Only] The **Artifact Type**
                for the Artifact.
            case_id (int, kwargs): [Required (alt: caseXid)] The **Case Id** for the Artifact.
            case_xid (str, kwargs): [Required (alt: caseId)] The **Case Xid** for the Artifact.
            date_added (str, kwargs): [Read-Only] The **Date Added** for the Artifact.
            file_data (str, kwargs): Base64 encoded file attachment required only for certain
                artifact types
            intel_type (str, kwargs): The **Intel Type** for the Artifact.
            links (Link, kwargs): The **Links** for the Artifact.
            notes (Note, kwargs): a list of Notes corresponding to the Artifact
            parent_case (Case, kwargs): [Read-Only] The **Parent Case** for the Artifact.
            source (str, kwargs): The **Source** for the Artifact.
            summary (str, kwargs): [Required] The **Summary** for the Artifact.
            task (case_management.task.Task, kwargs): [Read-Only] The **Task** for the Artifact.
            task_id (int, kwargs): the ID of the task which the Artifact references
            task_xid (str, kwargs): the XID of the task which the Artifact references
            type (str, kwargs): [Required] The **Type** for the Artifact.
        """
        return Artifact(self.tcex, **kwargs)

    def artifacts(self, **kwargs):
        """Return a instance of Artifacts object.

        .. code-block:: python
            :linenos:
            :lineno-start: 1

            # Example of params input
            {
                'result_limit': 100,  # How many results are retrieved.
                'result_start': 10,  # Starting point on retrieved results.
                'fields': ['caseId', 'summary']  # Additional fields returned on the results
            }

        Args:
            tcex (TcEx): An instantiated instance of TcEx object.
            initial_response (dict, optional): Initial data in
                Case Object for Artifact. Defaults to None.
            tql_filters (list, optional): List of TQL filters. Defaults to None.
            params(dict, optional): Dict of the params to be sent while
                retrieving the Artifacts objects.
        """
        return Artifacts(self.tcex, **kwargs)

    def artifact_type(self, **kwargs):
        """Return a instance of Artifact Type object.

        Args:
            tcex (TcEx): An instantiated instance of TcEx object.
            data_type (str, kwargs): [Read-Only] The **Data Type** for the Artifact Type.
            description (str, kwargs): [Read-Only] The **Description** for the Artifact Type.
            intel_type (str, kwargs): [Read-Only] The **Intel Type** for the Artifact Type.
            name (str, kwargs): [Read-Only] The **Name** for the Artifact Type.
        """
        return ArtifactType(self.tcex, **kwargs)

    def artifact_types(self, **kwargs):
        """Return a instance of Artifact Types object.

        .. code-block:: python
            :linenos:
            :lineno-start: 1

            # Example of params input
            {
                'result_limit': 100,  # How many results are retrieved.
                'result_start': 10,  # Starting point on retrieved results.
                'fields': ['caseId', 'summary']  # Additional fields returned on the results
            }

        Args:
            tcex (TcEx): An instantiated instance of TcEx object.
            initial_response (dict, optional): Initial data in
                Case Object for Artifact. Defaults to None.
            tql_filters (list, optional): List of TQL filters. Defaults to None.
            params(dict, optional): Dict of the params to be sent
                while retrieving the Artifact Type objects.
        """
        return ArtifactTypes(self.tcex, **kwargs)

    @staticmethod
    def assignee(**kwargs):
        """Return a instance of Assignee object.

        For User type the user_name or id fields is required and
        for the Group type the name or id is required.

        Args:
            first_name (str, kwargs): The first name of the User.
            id (id, kwargs): The id of the User.
            name (str, kwargs): The name of the Group.
            last_name (str, kwargs): The last name of the user.
            pseudonym (str, kwargs): The pseudonym of the User.
            role (str, kwargs): The role of the User.
            user_name (str, kwargs): The user name of the User.
            type (str, kwargs): The assignee type. Default to User.
        """
        return Assignee(**kwargs)

    def case(self, **kwargs):
        """Return a instance of Case object.

        Args:
            tcex (TcEx): An instantiated instance of TcEx object.
            artifacts (Artifact, kwargs): a list of Artifacts corresponding to the Case
            assignee (Assignee, kwargs): the user or group Assignee object for the Case
            created_by (User, kwargs): [Read-Only] The **Created By** for the Case.
            date_added (str, kwargs): [Read-Only] The **Date Added** for the Case.
            description (str, kwargs): The **Description** for the Case.
            name (str, kwargs): [Required] The **Name** for the Case.
            notes (Note, kwargs): a list of Notes corresponding to the Case
            owner (str, kwargs): The Org/Owner name for the Case.
            related (Case, kwargs): The **Related** for the Case.
            resolution (str, kwargs): The **Resolution** for the Case.
            severity (str, kwargs): [Required] The **Severity** for the Case.
            status (str, kwargs): [Required] The **Status** for the Case.
            tags (tcex.case_management.tag.Tag, kwargs): a list of Tags corresponding to the Case
                (NOTE: Setting this parameter will replace any existing tag(s) with
                the one(s) specified)
            tasks (case_management.task.Task, kwargs): a list of Tasks corresponding to the Case
            user_access (User, kwargs): a list of Users that, when defined, are the only
                ones allowed to view or edit the Case
            workflow_events (WorkflowEvent, kwargs): The **Events** for the Case.
            workflow_template (WorkflowTemplate, kwargs): the Template that the Case is
                populated by.
            xid (str, kwargs): The **Xid** for the Case.
        """
        return Case(self.tcex, **kwargs)

    def cases(self, **kwargs):
        """Return a instance of Cases object.

        .. code-block:: python
            :linenos:
            :lineno-start: 1

            # Example of params input
            {
                'result_limit': 100,  # How many results are retrieved.
                'result_start': 10,  # Starting point on retrieved results.
                'fields': ['caseId', 'summary']  # Additional fields returned on the results
            }

        Args:
            tcex (TcEx): An instantiated instance of TcEx object.
            initial_response (dict, optional): Initial data in
                Case Object for Artifact. Defaults to None.
            tql_filters (list, optional): List of TQL filters. Defaults to None.
            params(dict, optional): Dict of the params to be sent while
                retrieving the Case objects.
        """
        return Cases(self.tcex, **kwargs)

    def create_entity(self, entity, owner):
        """Create a CM object provided a dict and owner."""
        entity_type = entity.get('type').lower()
        entity_type = entity_type.replace(' ', '_')
        try:
            obj = getattr(self, entity_type)(**entity)
        except AttributeError:
            return None

        r = obj.submit()
        data = {'status_code': r.status_code}
        if r.ok:
            data.update(r.json().get('data', {}))
            data['main_type'] = 'Case_Management'
            data['sub_type'] = entity_type
            data['owner'] = owner

        return data

    def note(self, **kwargs):
        """Return a instance of Note object.

        Args:
            tcex (TcEx): An instantiated instance of TcEx object.
            artifact (Artifact, kwargs): [Read-Only] The **Artifact** for the Note.
            artifact_id (int, kwargs): the ID of the Artifact on which to apply the Note
            author (int, kwargs): [Read-Only] The **Author** for the Note.
            case_id (int, kwargs): [Required (alt: caseXid)] The **Case Id** for the Note.
            case_xid (str, kwargs): [Required (alt: caseId)]
            date_added (str, kwargs): [Read-Only] The **Date Added** for the Note.
            edited (boolean, kwargs): [Read-Only] The **Edited** for the Note.
            last_modified (str, kwargs): [Read-Only] The **Last Modified** for the Note.
            parent_case (Case, kwargs): [Read-Only] The **Parent Case** for the Note.
            summary (str, kwargs): [Read-Only] The **Summary** for the Note.
            task (case_management.task.Task, kwargs): [Read-Only] The **Task** for the Note.
            task_id (int, kwargs): the ID of the Task on which to apply the Note
            task_xid (str, kwargs): the XID of the Task on which to apply the Note
            text (str, kwargs): [Required] The **Text** for the Note.
            workflow_event (WorkflowEvent, kwargs): [Read-Only] The **Workflow Event** for the Note.
            workflow_event_id (int, kwargs): the ID of the Event on which to apply the Note
        """
        return Note(self.tcex, **kwargs)

    def notes(self, **kwargs):
        """Return a instance of Notes object.

        .. code-block:: python
            :linenos:
            :lineno-start: 1

            # Example of params input
            {
                'result_limit': 100,  # How many results are retrieved.
                'result_start': 10,  # Starting point on retrieved results.
                'fields': ['caseId', 'summary']  # Additional fields returned on the results
            }

        Args:
            tcex (TcEx): An instantiated instance of TcEx object.
            initial_response (dict, optional): Initial data in
                Case Object for Note. Defaults to None.
            tql_filters (list, optional): List of TQL filters. Defaults to None.
            params(dict, optional): Dict of the params to be sent while
                retrieving the Notes objects.
        """
        return Notes(self.tcex, **kwargs)

    def tag(self, **kwargs):
        """Return a instance of Tag object.

        Args:
            tcex (TcEx): An instantiated instance of TcEx object.
            cases (Case, kwargs): [Read-Only] The **Cases** for the Tag.
            description (str, kwargs): a brief description of the Tag
            last_used (str, kwargs): [Read-Only] The **Last Used** for the Tag.
            name (str, kwargs): [Required] The **Name** for the Tag.
            owner (str, kwargs): [Read-Only] The **Owner** for the Tag.
        """
        return Tag(self.tcex, **kwargs)

    def tags(self, **kwargs):
        """Return a instance of Tags object.

        .. code-block:: python
            :linenos:
            :lineno-start: 1

            # Example of params input
            {
                'result_limit': 100,  # How many results are retrieved.
                'result_start': 10,  # Starting point on retrieved results.
                'fields': ['caseId', 'summary']  # Additional fields returned on the results
            }

        Args:
            tcex (TcEx): An instantiated instance of TcEx object.
            initial_response (dict, optional): Initial data in
                Case Object for Tag. Defaults to None.
            tql_filters (list, optional): List of TQL filters. Defaults to None.
            params(dict, optional): Dict of the params to be sent while
                retrieving the Tag objects.
        """
        return Tags(self.tcex, **kwargs)

    def task(self, **kwargs):
        """Return a instance of Task object.

        Args:
            tcex (TcEx): An instantiated instance of TcEx object.
            artifacts (Artifact, kwargs): a list of Artifacts corresponding to the Task
            assignee (Assignee, kwargs): the user or group Assignee object for the Task
            case_id (int, kwargs): [Required (alt: caseXid)] The **Case Id** for the Task.
            case_xid (str, kwargs): [Required (alt: caseId)] The **Case Xid** for the Task.
            completed_by (str, kwargs): [Read-Only] The **Completed By** for the Task.
            completed_date (str, kwargs): the completion date of the Task
            config_playbook (str, kwargs): [Read-Only] The **Config Playbook** for the Task.
            config_task (dict, kwargs): [Read-Only] The **Config Task** for the Task.
            dependent_on_id (str, kwargs): [Read-Only] The **Dependent On ID** for the Task.
            description (str, kwargs): The **Description** for the Task.
            due_date (str, kwargs): the due date of the Task
            duration (int, kwargs): [Read-Only] The **Duration** for the Task.
            id_dependent_on (int, kwargs): [Read-Only] The **Id_Dependent On** for the Task.
            name (str, kwargs): [Required] The **Name** for the Task.
            notes (Note, kwargs): a list of Notes corresponding to the Task
            parent_case (Case, kwargs): [Read-Only] The **Parent Case** for the Task.
            required (bool, kwargs): [Read-Only] The **Required** flag for the Task.
            status (str, kwargs): The **Status** for the Task.
            workflow_phase (int, kwargs): the phase of the workflow
            workflow_step (int, kwargs): the step of the workflow
            xid (str, kwargs): The **Xid** for the Task.
        """
        return Task(self.tcex, **kwargs)

    def tasks(self, **kwargs):
        """Return a instance of Tasks object.

        .. code-block:: python
            :linenos:
            :lineno-start: 1

            # Example of params input
            {
                'result_limit': 100,  # How many results are retrieved.
                'result_start': 10,  # Starting point on retrieved results.
                'fields': ['caseId', 'summary']  # Additional fields returned on the results
            }

        Args:
            tcex (TcEx): An instantiated instance of TcEx object.
            initial_response (dict, optional): Initial data in
                Case Object for Task. Defaults to None.
            tql_filters (list, optional): List of TQL filters. Defaults to None.
            params(dict, optional): Dict of the params to be sent while
                retrieving the Task objects.
        """
        return Tasks(self.tcex, **kwargs)

    @staticmethod
    def user(**kwargs):
        """Return a instance of User object.

        Args:
            first_name (str, kwargs): The first name of the user.
            id (id, kwargs): The id of the user.
            last_name (str, kwargs): The last name of the user.
            pseudonym (str, kwargs): The pseudonym of the user.
            role (str, kwargs): The role of the user.
            user_name (str, kwargs): The user name of the user.
        """
        return User(**kwargs)

    @staticmethod
    def users(users):
        """Sub class of the Cases object. Used to map the users to.

        Args:
            users (list): A array of user data
        """
        return Users(users)

    def workflow_event(self, **kwargs):
        """Return a instance of Workflow Event object.

        Args:
            tcex (TcEx): An instantiated instance of TcEx object.
            case_id (int, kwargs): [Required (alt: caseXid)] The **Case Id** for the Workflow Event.
            case_xid (str, kwargs): [Required (alt: caseId)] The **Case Xid**
                for the Workflow Event.
            date_added (str, kwargs): [Read-Only] The **Date Added** for the Workflow Event.
            deleted (bool, kwargs): [Read-Only] The **Deleted** flag for the Workflow Event.
            deleted_reason (str, kwargs): the reason for deleting the event (required
                input for DELETE operation only)
            event_date (str, kwargs): the time that the Event is logged
            link (str, kwargs): [Read-Only] The **Link** for the Workflow Event.
            link_text (str, kwargs): [Read-Only] The **Link Text** for the Workflow Event.
            notes (Note, kwargs): a list of Notes corresponding to the Event
            parent_case (Case, kwargs): [Read-Only] The **Parent Case** for the Workflow Event.
            summary (str, kwargs): [Required] The **Summary** for the Workflow Event.
            system_generated (bool, kwargs): [Read-Only] The **System Generated**
                flag for the Workflow Event.
            user (User, kwargs): [Read-Only] The **User** for the Workflow Event.
        """
        return WorkflowEvent(self.tcex, **kwargs)

    def workflow_events(self, **kwargs):
        """Return a instance of Workflow Events object.

        .. code-block:: python
            :linenos:
            :lineno-start: 1

            # Example of params input
            {
                'result_limit': 100,  # How many results are retrieved.
                'result_start': 10,  # Starting point on retrieved results.
                'fields': ['caseId', 'summary']  # Additional fields returned on the results
            }

        Args:
            tcex (TcEx): An instantiated instance of TcEx object.
            initial_response (dict, optional): Initial data in
                Case Object for Artifact. Defaults to None.
            tql_filters (list, optional): List of TQL filters. Defaults to None.
            params(dict, optional): Dict of the params to be sent while
                retrieving the Workflow Event objects.
        """
        return WorkflowEvents(self.tcex, **kwargs)

    def workflow_template(self, **kwargs):
        """Return a instance of Workflow Template object.

        Args:
            tcex (TcEx): An instantiated instance of TcEx object.
            active (bool, kwargs): [Read-Only] The **Active** flag for the Workflow Template.
            assignee (Assignee, kwargs): [Read-Only] The **Assignee** for the Workflow Template.
            cases (Case, kwargs): [Read-Only] The **Cases** for the Workflow Template.
            config_artifact (str, kwargs): [Read-Only] The **Config Artifact** for the Workflow
                Template.
            config_playbook (str, kwargs): [Read-Only] The **Config Playbook** for the Workflow
                Template.
            config_task (dict, kwargs): [Read-Only] The **Config Task** for the Workflow Template.
            description (str, kwargs): The **Description** for the Workflow Template.
            name (str, kwargs): [Required] The **Name** for the Workflow Template.
            target_type (str, kwargs): [Read-Only] The **Target Type** for the Workflow Template.
            version (int, kwargs): The **Version** for the Workflow Template.
        """
        return WorkflowTemplate(self.tcex, **kwargs)

    def workflow_templates(self, **kwargs):
        """Return a instance of Workflow Templates object.

        .. code-block:: python
            :linenos:
            :lineno-start: 1

            # Example of params input
            {
                'result_limit': 100,  # How many results are retrieved.
                'result_start': 10,  # Starting point on retrieved results.
                'fields': ['caseId', 'summary']  # Additional fields returned on the results
            }

        Args:
            tcex (TcEx): An instantiated instance of TcEx object.
            initial_response (dict, optional): Initial data in
                Case Object for Workflow Template. Defaults to None.
            tql_filters (list, optional): List of TQL filters. Defaults to None.
            params(dict, optional): Dict of the params to be sent while
                retrieving the Workflow Event objects.
        """
        return WorkflowTemplates(self.tcex, **kwargs)



"""App Decorators Module."""
# standard library
import datetime

# third-party
import wrapt


class Benchmark:
    """Log benchmarking times.

    This decorator will log the time of execution (benchmark_time) to the app.log file. It can be
    helpful in troubleshooting performance issues with Apps.

    .. code-block:: python
        :linenos:
        :lineno-start: 1

        import time

        @Benchmark()
        def my_method():
            time.sleep(1)
    """

    @wrapt.decorator
    def __call__(self, wrapped, instance, args, kwargs):
        """Implement __call__ function for decorator.

        Args:
            wrapped (callable): The wrapped function which in turns
                needs to be called by your wrapper function.
            instance (App): The object to which the wrapped
                function was bound when it was called.
            args (list): The list of positional arguments supplied
                when the decorated function was called.
            kwargs (dict): The dictionary of keyword arguments
                supplied when the decorated function was called.

        Returns:
            function: The custom decorator function.
        """

        def benchmark(app, *args, **kwargs):
            """Iterate over data, calling the decorated function for each value.

            Args:
                app (class): The instance of the App class "self".
            """

            before = datetime.datetime.now()
            data = wrapped(*args, **kwargs)
            after = datetime.datetime.now()
            app.tcex.log.debug(
                f'function: "{self.__class__.__name__}", benchmark_time: "{after - before}"'
            )
            return data

        return benchmark(instance, *args, **kwargs)



#!/usr/bin/env python
"""TcEx Framework Validate Module."""
# standard library
import ast
import importlib
import json
import os
import sys
import traceback
from collections import deque

# third-party
import colorama as c
from jsonschema import SchemaError, ValidationError, validate
from stdlib_list import stdlib_list

from .bin import Bin

try:
    # third-party
    import pkg_resources
except PermissionError:
    # this module is only required for certain CLI commands
    pass

try:
    # standard library
    import sqlite3
except ModuleNotFoundError:
    # this module is only required for certain CLI commands
    pass


class Validate(Bin):
    """Validate syntax, imports, and schemas.

    * Python and JSON file syntax
    * Python import modules
    * install.json schema
    * layout.json schema

    Args:
        _args (namespace): The argparser args Namespace.
    """

    def __init__(self, _args):
        """Init Class properties.

        Args:
            _args (namespace): The argparser args Namespace.
        """
        super().__init__(_args)

        # class properties
        self._app_packages = []
        self._install_json_schema = None
        self._layout_json_schema = None
        self.config = {}

        if 'pkg_resources' in sys.modules:
            # only set these if pkg_resource module is available
            pkg_path = pkg_resources.resource_filename(__name__, '').rstrip('bin')
            self.install_json_schema_file = os.path.join(
                pkg_path, 'schemas', 'install-json-schema.json'
            )
            self.layout_json_schema_file = os.path.join(
                pkg_path, 'schemas', 'layout-json-schema.json'
            )
        else:
            self.install_json_schema_file = None
            self.layout_json_schema_file = None
        self.validation_data = self._validation_data

    @property
    def _validation_data(self):
        """Return structure for validation data."""
        return {'errors': [], 'fileSyntax': [], 'layouts': [], 'moduleImports': [], 'schema': []}

    def check_imports(self):
        """Check the projects top level directory for missing imports.

        This method will check only files ending in **.py** and does not handle imports validation
        for sub-directories.
        """
        modules = []
        for filename in sorted(os.listdir(self.app_path)):
            if not filename.endswith('.py'):
                continue

            fq_path = os.path.join(self.app_path, filename)
            with open(fq_path, 'rb') as f:
                # TODO: fix this
                code_lines = deque([(f.read(), 1)])

                while code_lines:
                    m_status = True
                    code, lineno = code_lines.popleft()  # pylint: disable=unused-variable
                    try:
                        parsed_code = ast.parse(code)
                        for node in ast.walk(parsed_code):
                            if isinstance(node, ast.Import):
                                for n in node.names:
                                    m = n.name.split('.')[0]
                                    if self.check_import_stdlib(m):
                                        # stdlib module, not need to proceed
                                        continue
                                    m_status = self.check_imported(m)
                                    modules.append(
                                        {'file': filename, 'module': m, 'status': m_status}
                                    )
                            elif isinstance(node, ast.ImportFrom):
                                m = node.module.split('.')[0]
                                if self.check_import_stdlib(m):
                                    # stdlib module, not need to proceed
                                    continue
                                m_status = self.check_imported(m)
                                modules.append({'file': filename, 'module': m, 'status': m_status})
                            else:
                                continue
                    except SyntaxError:
                        pass

        for module_data in modules:
            status = True
            if not module_data.get('status'):
                status = False
                # update validation data errors
                self.validation_data['errors'].append(
                    f"""Module validation failed for {module_data.get('file')} """
                    f"""(module "{module_data.get('module')}" could not be imported)."""
                )
            # update validation data for module
            self.validation_data['moduleImports'].append(
                {
                    'filename': module_data.get('file'),
                    'module': module_data.get('module'),
                    'status': status,
                }
            )

    @staticmethod
    def check_import_stdlib(module):
        """Check if module is in Python stdlib.

        Args:
            module (str): The name of the module to check.

        Returns:
            bool: Returns True if the module is in the stdlib or template.
        """
        if (
            module in stdlib_list('3.6')
            or module in stdlib_list('3.7')
            or module in stdlib_list('3.8')
            or module in ['app', 'args', 'job_app', 'playbook_app', 'run', 'service_app']
        ):
            return True
        return False

    @staticmethod
    def check_imported(module):
        """Check whether the provide module can be imported (package installed).

        Args:
            module (str): The name of the module to check availability.

        Returns:
            bool: True if the module can be imported, False otherwise.
        """
        try:
            del sys.modules[module]
        except (AttributeError, KeyError):
            pass

        # https://docs.python.org/3/library/importlib.html#checking-if-a-module-can-be-imported
        find_spec = importlib.util.find_spec(module)
        found = find_spec is not None
        if found is True:
            # if dist-packages|site-packages in module_path the import doesn't count
            if 'dist-packages' in find_spec.origin:
                found = False
            if 'site-packages' in find_spec.origin:
                found = False
        return found

    def check_install_json(self):
        """Check all install.json files for valid schema."""
        if self.install_json_schema is None:
            return

        contents = os.listdir(self.app_path)
        if self.args.install_json is not None:
            contents = [self.args.install_json]

        for install_json in sorted(contents):
            # skip files that are not install.json files
            if 'install.json' not in install_json:
                continue

            error = None
            status = True

            try:
                # loading explicitly here to keep all error catching in this file
                with open(install_json) as fh:
                    data = json.loads(fh.read())
                validate(data, self.install_json_schema)
            except SchemaError as e:
                status = False
                error = e
            except ValidationError as e:
                status = False
                error = e.message
            except ValueError:
                # any JSON decode error will be caught during syntax validation
                return

            if error:
                # update validation data errors
                self.validation_data['errors'].append(
                    f'Schema validation failed for {install_json} ({error}).'
                )

            # update validation data for module
            self.validation_data['schema'].append({'filename': install_json, 'status': status})

    def check_layout_json(self):
        """Check all layout.json files for valid schema."""
        # the install.json files can't be validates if the schema file is not present
        layout_json_file = 'layout.json'
        if self.layout_json_schema is None or not os.path.isfile(layout_json_file):
            return

        error = None
        status = True
        try:
            # loading explicitly here to keep all error catching in this file
            with open(layout_json_file) as fh:
                data = json.loads(fh.read())
            validate(data, self.layout_json_schema)
        except SchemaError as e:
            status = False
            error = e
        except ValidationError as e:
            status = False
            error = e.message
        except ValueError:
            # any JSON decode error will be caught during syntax validation
            return

        # update validation data for module
        self.validation_data['schema'].append({'filename': layout_json_file, 'status': status})

        if error:
            # update validation data errors
            self.validation_data['errors'].append(
                f'Schema validation failed for {layout_json_file} ({error}).'
            )
        else:
            self.check_layout_params()

    def check_layout_params(self):
        """Check that the layout.json is consistent with install.json.

        The layout.json files references the params.name from the install.json file.  The method
        will validate that no reference appear for inputs in install.json that don't exist.
        """

        # do not track hidden or serviceConfig inputs as they should not be in layouts.json
        ij_input_names = [
            p.get('name')
            for p in self.ij.filter_params_dict(service_config=False, hidden=False).values()
        ]
        ij_output_names = [o.get('name') for o in self.ij.output_variables]

        # Check for duplicate inputs
        for name in self.ij.validate_duplicate_input_names():
            self.validation_data['errors'].append(
                f'Duplicate input name found in install.json ({name})'
            )
            status = False

        # Check for duplicate sequence numbers
        for sequence in self.ij.validate_duplicate_sequences():
            self.validation_data['errors'].append(
                f'Duplicate sequence number found in install.json ({sequence})'
            )
            status = False

        # Check for duplicate outputs variables
        for output in self.ij.validate_duplicate_outputs():
            self.validation_data['errors'].append(
                f'Duplicate output variable name found in install.json ({output})'
            )
            status = False

        if 'sqlite3' in sys.modules:
            # create temporary inputs tables
            self.permutations.db_create_table(self.permutations.input_table, ij_input_names)

        # inputs
        status = True
        for i in self.lj.inputs:
            for p in i.get('parameters'):
                if p.get('name') not in ij_input_names:
                    # update validation data errors
                    self.validation_data['errors'].append(
                        'Layouts input.parameters[].name validations failed '
                        f"""("{p.get('name')}" is defined in layout.json, """
                        'but hidden or not found in install.json).'
                    )
                    status = False
                else:
                    # any item in list afterwards is a problem
                    ij_input_names.remove(p.get('name'))

                if 'sqlite3' in sys.modules:
                    if p.get('display'):
                        display_query = (
                            f'''SELECT * FROM {self.permutations.input_table}'''  # nosec
                            f''' WHERE {p.get('display')}'''
                        )
                        try:
                            self.permutations.db_conn.execute(display_query.replace('"', ''))
                        except sqlite3.Error:
                            self.validation_data['errors'].append(
                                '''Layouts input.parameters[].display validations failed '''
                                f'''("{p.get('display')}" query is an invalid statement).'''
                            )
                            status = False

        # update validation data for module
        self.validation_data['layouts'].append({'params': 'inputs', 'status': status})

        if ij_input_names:
            input_names = ','.join(ij_input_names)
            # update validation data errors
            self.validation_data['errors'].append(
                f'Layouts input.parameters[].name validations failed ("{input_names}" '
                'values from install.json were not included in layout.json.'
            )
            status = False

        # outputs
        status = True
        for o in self.lj.outputs:
            if o.get('name') not in ij_output_names:
                # update validation data errors
                self.validation_data['errors'].append(
                    f'''Layouts output validations failed ({o.get('name')} is defined '''
                    '''in layout.json, but not found in install.json).'''
                )
                status = False

            if 'sqlite3' in sys.modules:
                if o.get('display'):
                    display_query = (
                        f'''SELECT * FROM {self.permutations.input_table} '''  # nosec
                        f'''WHERE {o.get('display')}'''
                    )
                    try:
                        self.permutations.db_conn.execute(display_query.replace('"', ''))
                    except sqlite3.Error:
                        self.validation_data['errors'].append(
                            f"""Layouts outputs.display validations failed ("{o.get('display')}" """
                            f"""query is an invalid statement)."""
                        )
                        status = False

        # update validation data for module
        self.validation_data['layouts'].append({'params': 'outputs', 'status': status})

    def check_syntax(self, app_path=None):
        """Run syntax on each ".py" and ".json" file.

        Args:
            app_path (str, optional): Defaults to None. The path of Python files.
        """
        app_path = app_path or '.'

        for filename in sorted(os.listdir(app_path)):
            error = None
            status = True
            if filename.endswith('.py'):
                try:
                    with open(filename, 'rb') as f:
                        ast.parse(f.read(), filename=filename)
                except SyntaxError:
                    status = False
                    # cleanup output
                    e = []
                    for line in traceback.format_exc().split('\n')[-5:-2]:
                        e.append(line.strip())
                    error = ' '.join(e)

            elif filename.endswith('.json'):
                try:
                    with open(filename) as fh:
                        json.load(fh)
                except ValueError as e:
                    status = False
                    error = e
            else:
                # skip unsupported file types
                continue

            if error:
                # update validation data errors
                self.validation_data['errors'].append(
                    f'Syntax validation failed for {filename} ({error}).'
                )

            # store status for this file
            self.validation_data['fileSyntax'].append({'filename': filename, 'status': status})

    @property
    def install_json_schema(self):
        """Load install.json schema file."""
        if self._install_json_schema is None and self.install_json_schema_file is not None:
            # remove old schema file
            if os.path.isfile('tcex_json_schema.json'):
                # this file is now part of tcex.
                os.remove('tcex_json_schema.json')

            if os.path.isfile(self.install_json_schema_file):
                with open(self.install_json_schema_file) as fh:
                    self._install_json_schema = json.load(fh)
        return self._install_json_schema

    def interactive(self):
        """Run in interactive mode."""
        while True:
            line = sys.stdin.readline().strip()
            if line == 'quit':
                sys.exit()
            elif line == 'validate':
                self.check_syntax()
                self.check_imports()
                self.check_install_json()
                self.check_layout_json()
                self.print_json()

            # reset validation_data
            self.validation_data = self._validation_data

    @property
    def layout_json_schema(self):
        """Load layout.json schema file."""
        if self._layout_json_schema is None and self.layout_json_schema_file is not None:
            if os.path.isfile(self.layout_json_schema_file):
                with open(self.layout_json_schema_file) as fh:
                    self._layout_json_schema = json.load(fh)
        return self._layout_json_schema

    def print_json(self):
        """Print JSON output."""
        print(json.dumps({'validation_data': self.validation_data}))

    def print_results(self):
        """Print results."""
        # Validating Syntax
        if self.validation_data.get('fileSyntax'):
            print(f'\n{c.Style.BRIGHT}{c.Fore.BLUE}Validated File Syntax:')
            print(f"{c.Style.BRIGHT}{'File:'!s:<60}{'Status:'!s:<25}")
            for f in self.validation_data.get('fileSyntax'):
                status_color = self.status_color(f.get('status'))
                status_value = self.status_value(f.get('status'))
                print(f"{f.get('filename')!s:<60}{status_color}{status_value!s:<25}")

        # Validating Imports
        if self.validation_data.get('moduleImports'):
            print(f'\n{c.Style.BRIGHT}{c.Fore.BLUE}Validated Imports:')
            print(f"{c.Style.BRIGHT}{'File:'!s:<30}{'Module:'!s:<30}{'Status:'!s:<25}")
            for f in self.validation_data.get('moduleImports'):
                status_color = self.status_color(f.get('status'))
                status_value = self.status_value(f.get('status'))
                print(
                    f"{f.get('filename')!s:<30}{c.Fore.WHITE}"
                    f"{f.get('module')!s:<30}{status_color}{status_value!s:<25}"
                )

        # Validating Schema
        if self.validation_data.get('schema'):
            print(f'\n{c.Style.BRIGHT}{c.Fore.BLUE}Validated Schema:')
            print(f"{c.Style.BRIGHT}{'File:'!s:<60}{'Status:'!s:<25}")
            for f in self.validation_data.get('schema'):
                status_color = self.status_color(f.get('status'))
                status_value = self.status_value(f.get('status'))
                print(f"{f.get('filename')!s:<60}{status_color}{status_value!s:<25}")

        # Validating Layouts
        if self.validation_data.get('layouts'):
            print(f'\n{c.Style.BRIGHT}{c.Fore.BLUE}Validated Layouts:')
            print(f"{c.Style.BRIGHT}{'Params:'!s:<60}{'Status:'!s:<25}")
            for f in self.validation_data.get('layouts'):
                status_color = self.status_color(f.get('status'))
                status_value = self.status_value(f.get('status'))
                print(f"{f.get('params')!s:<60}{status_color}{status_value!s:<25}")

        if self.validation_data.get('errors'):
            print('\n')  # separate errors from normal output
        for error in self.validation_data.get('errors'):
            # print all errors
            print(f'* {c.Fore.RED}{error}')

            # ignore exit code
            if not self.args.ignore_validation:
                self.exit_code = 1

    @staticmethod
    def status_color(status):
        """Return the appropriate status color."""
        status_color = c.Fore.GREEN
        if not status:
            status_color = c.Fore.RED
        return status_color

    @staticmethod
    def status_value(status):
        """Return the appropriate status color."""
        status_value = 'passed'
        if not status:
            status_value = 'failed'
        return status_value



"""TcEx Framework Logger module"""
# standard library
import logging
import os
import pathlib
import platform
import sys
from typing import Optional

from .api_handler import ApiHandler, ApiHandlerFormatter
from .cache_handler import CacheHandler
from .pattern_file_handler import PatternFileHandler
from .rotating_file_handler_custom import RotatingFileHandlerCustom
from .thread_file_handler import ThreadFileHandler
from .trace_logger import TraceLogger


class Logger:
    """Framework logger module."""

    def __init__(self, tcex: object, logger_name: str):
        """Initialize Class Properties.

        Args:
            tcex: Instance of TcEx class.
            logger_name: The name of the logger.
        """
        self.tcex = tcex
        self.logger_name = logger_name

    @property
    def _logger(self) -> logging.Logger:
        """Return the logger. The default_args property is not available in init."""
        logging.setLoggerClass(TraceLogger)
        logger = logging.getLogger(self.logger_name)
        logger.setLevel(logging.TRACE)
        return logger

    @property
    def _formatter(self) -> None:
        """Return log formatter."""
        tx_format = (
            '%(asctime)s - %(name)s - %(levelname)8s - %(message)s '
            '(%(filename)s:%(funcName)s:%(lineno)d)'
        )
        return logging.Formatter(tx_format)

    @property
    def _formatter_thread_name(self) -> None:
        """Return log formatter."""
        tx_format = (
            '%(asctime)s - %(name)s - %(levelname)8s - %(message)s '
            '(%(filename)s:%(funcName)s:%(lineno)d:%(threadName)s)'
        )
        return logging.Formatter(tx_format)

    def handler_exist(self, handler_name: str) -> bool:
        """Remove a file handler by name.

        Args:
            handler_name: The handler name to remove.

        Returns:
            bool: True if handler current exists
        """
        for h in self._logger.handlers:
            if h.get_name() == handler_name:
                return True
        return False

    @property
    def log(self) -> logging.Logger:
        """Return logger."""
        return self._logger

    @staticmethod
    def log_level(level: str) -> int:
        """Return proper level from string.

        Args:
            level: The logging level.

        Returns:
            int: The logging level as an int.
        """
        level = level or 'debug'
        return logging.getLevelName(level.upper())

    def remove_handler_by_name(self, handler_name: str) -> None:
        """Remove a file handler by name.

        Args:
            handler_name: The handler name to remove.
        """
        for h in self._logger.handlers:
            if h.get_name() == handler_name:
                self._logger.removeHandler(h)
                break

    def replay_cached_events(self, handler_name: Optional[str] = 'cache') -> None:
        """Replay cached log events and remove handler."""
        for h in self._logger.handlers:
            if h.get_name() == handler_name:
                events = h.events
                self._logger.removeHandler(h)
                for event in events:
                    self._logger.handle(event)
                break

    def shutdown(self) -> None:
        """Close all handlers.

        Args:
            handler_name (str): The handler name to remove.
        """
        for h in self._logger.handlers:
            self._logger.removeHandler(h)

    def update_handler_level(self, level: int) -> None:
        """Update all handlers log level.

        Args:
            level: The logging level.
        """
        level = self.log_level(level)

        # update all handler logging levels
        for h in self._logger.handlers:
            h.setLevel(level)

    #
    # handlers
    #

    def add_api_handler(self, name: Optional[str] = 'api', level: Optional[str] = None) -> None:
        """Add API logging handler.

        Args:
            name: The name of the handler.
            level: The level value as a string.
        """
        self.remove_handler_by_name(name)
        api = ApiHandler(self.tcex.session)
        api.set_name(name)
        api.setLevel(self.log_level(level))
        api.setFormatter(ApiHandlerFormatter())
        self._logger.addHandler(api)

    def add_cache_handler(self, name: str) -> None:
        """Add cache logging handler.

        Args:
            name: The name of the handler.
        """
        self.remove_handler_by_name(name)
        cache = CacheHandler()
        cache.set_name(name)
        # set logging level to INFO as event will typically
        # be only those that happen before args are processed
        cache.setLevel(self.log_level('trace'))
        cache.setFormatter(self._formatter)
        self._logger.addHandler(cache)

    def add_pattern_file_handler(
        self,
        name: str,
        filename: str,
        level: int,
        path: str,
        pattern: str,
        formatter: Optional[str] = None,
        handler_key: Optional[str] = None,
        max_log_count: Optional[int] = 100,
        thread_key: Optional[str] = None,
    ) -> None:
        """Add custom file logging handler.

        This handler is intended for service Apps that need to log events based on the
        current session id. All log event would be in context to a single playbook execution.

        Args:
            name: The name of the handler.
            filename: The name of the logfile.
            level: The logging level.
            path: The path for the logfile.
            formatter: The logging formatter to use.
            handler_key: Additional properties for handler to thread condition.
            max_log_count: The maximun number of logs to keep that match the provided pattern.
            pattern: The pattern used to match the log files.
            thread_key: Additional properties for handler to thread condition.
        """
        self.remove_handler_by_name(name)
        formatter = formatter or self._formatter
        # create customized handler
        fh = PatternFileHandler(
            filename=os.path.join(path, filename), pattern=pattern, max_log_count=max_log_count
        )
        fh.set_name(name)
        fh.setFormatter(formatter)
        fh.setLevel(self.log_level(level))
        # add keys for halder emit method conditional
        fh.handler_key = handler_key
        fh.thread_key = thread_key
        self._logger.addHandler(fh)

    def add_rotating_file_handler(
        self,
        name: str,
        filename: str,
        path: str,
        backup_count: int,
        max_bytes: int,
        level: str,
        formatter: Optional[str] = None,
        mode: Optional[str] = 'a',
    ) -> None:
        """Add custom file logging handler.

        Args:
            name: The name of the handler.
            filename: The name of the logfile.
            path: The path for the logfile.
            backup_count: The maximum # of backup files.
            max_bytes: The max file size before rotating.
            level: The logging level.
            formatter: The logging formatter to use.
            mode: The write mode for the file.
        """
        self.remove_handler_by_name(name)
        formatter = formatter or self._formatter_thread_name

        # create customized handler
        fh = RotatingFileHandlerCustom(
            os.path.join(path, filename), backupCount=backup_count, maxBytes=max_bytes, mode=mode
        )
        fh.set_name(name)
        fh.setFormatter(formatter)
        fh.setLevel(self.log_level(level))
        self._logger.addHandler(fh)

    def add_stream_handler(
        self,
        name: Optional[str] = 'sh',
        formatter: Optional[str] = None,
        level: Optional[int] = None,
    ) -> None:
        """Add stream logging handler.

        Args:
            name: The name of the handler.
            formatter: The logging formatter to use.
            level: The logging level.
        """
        self.remove_handler_by_name(name)
        formatter = formatter or self._formatter
        # create handler
        sh = logging.StreamHandler()
        sh.set_name(name)
        sh.setFormatter(formatter)
        sh.setLevel(self.log_level(level))
        self._logger.addHandler(sh)

    def add_thread_file_handler(
        self,
        name: str,
        filename: str,
        level: str,
        path: str,
        backup_count: Optional[int] = 0,
        formatter: Optional[str] = None,
        handler_key: Optional[str] = None,
        max_bytes: Optional[int] = 0,
        mode: Optional[str] = 'a',
        thread_key: Optional[str] = None,
    ) -> None:
        """Add custom file logging handler.

        This handler is intended for service Apps that need to log events based on the
        current trigger id. All log events would be in context to a single playbook.

        Args:
            name: The name of the handler.
            filename: The name of the logfile.
            level: The logging level.
            path: The path for the logfile.
            backup_count: The maximum # of backup files.
            formatter: The logging formatter to use.
            handler_key: Additional properties for handler to thread condition.
            max_bytes: The max file size before rotating.
            mode: The write mode for the file.
            thread_key: Additional properties for handler to thread condition.
        """
        self.remove_handler_by_name(name)
        formatter = formatter or self._formatter_thread_name
        # create customized handler
        fh = ThreadFileHandler(
            os.path.join(path, filename), backupCount=backup_count, maxBytes=max_bytes, mode=mode
        )
        fh.set_name(name)
        fh.setFormatter(formatter)
        fh.setLevel(self.log_level(level))
        # add keys for halder emit method conditional
        fh.handler_key = handler_key
        fh.thread_key = thread_key
        self._logger.addHandler(fh)

    #
    # App info logging
    #

    def log_info(self, args: object) -> None:
        """Send System and App data to logs.

        Args:
            args: The args namespace for App.
        """
        self._log_platform()
        self._log_app_data()
        self._log_python_version()
        self._log_tcex_version()
        self._log_tcex_path()
        self._log_tc_proxy(args)

    def _log_app_data(self) -> None:
        """Log the App data information as a best effort."""
        try:
            self.log.info(f'app-name="{self.tcex.ij.display_name}"')
            if self.tcex.ij.features:
                self.log.info(f'app-features={self.tcex.ij.features}')
            self.log.info(f'app-minimum-threatconnect-version={self.tcex.ij.min_server_version}')
            self.log.info(f'app-runtime-level={self.tcex.ij.runtime_level}')
            self.log.info(f'app-version={self.tcex.ij.program_version}')
            if self.tcex.ij.commit_hash is not None:
                self.log.info(f'app-commit-hash={self.tcex.ij.commit_hash}')
        except Exception:  # nosec; pragma: no cover
            pass

    def _log_platform(self) -> None:
        """Log the current Platform."""
        self.log.info(f'platform="{platform.platform()}"')

    def _log_python_version(self) -> None:
        """Log the current Python version."""
        self.log.info(
            f'python-version={sys.version_info.major}.'
            f'{sys.version_info.minor}.'
            f'{sys.version_info.micro}'
        )

    def _log_tc_proxy(self, args: object) -> None:
        """Log the proxy settings.

        Args:
            args: The args namespace for App.
        """
        if args.tc_proxy_tc:
            self.log.info(f'proxy-server-tc={args.tc_proxy_host}:{args.tc_proxy_port}')

    def _log_tcex_version(self) -> None:
        """Log the current TcEx version number."""
        self.log.info(f'tcex-version={__import__(__name__).__version__}')

    def _log_tcex_path(self) -> None:
        """Log the current TcEx path."""
        app_path = str(pathlib.Path().parent.absolute())
        full_path = str(pathlib.Path(__file__).parent.absolute())
        tcex_path = os.path.dirname(full_path.replace(app_path, ''))
        self.log.info(f'tcex-path="{tcex_path}"')



"""App Decorators Module."""
# third-party
import wrapt


class Output:
    """Store the method return value in self.<attribute>.

    This decorator will write, append, or extend the methods return value to the App attribute
    provided in the ``attribute`` input. The ``attribute`` must first be defined in the
    ``__init__()`` method of the App before the decorator is used.

    .. code-block:: python
        :linenos:
        :lineno-start: 1

        def __init__(self, _tcex):
            super(App, self).__init__(_tcex)
            self.output_strings = []  # Output decorator writes here.

        @Output(attribute='output_strings')
        def my_method(data):
            return data.lowercase()

    Args:
        attribute (str): The name of the App attribute to write data.
        overwrite (bool): When True and the method is called more than once the previous value
            will be overwritten.
    """

    def __init__(self, attribute, overwrite=False):
        """Initialize Class properties"""
        self.attribute = attribute
        self.overwrite = overwrite

    @wrapt.decorator
    def __call__(self, wrapped, instance, args, kwargs):
        """Implement __call__ function for decorator.

        Args:
            wrapped (callable): The wrapped function which in turns
                needs to be called by your wrapper function.
            instance (App): The object to which the wrapped
                function was bound when it was called.
            args (list): The list of positional arguments supplied
                when the decorated function was called.
            kwargs (dict): The dictionary of keyword arguments
                supplied when the decorated function was called.

        Returns:
            function: The custom decorator function.
        """

        def output(app, *args, **kwargs):
            """Call the function and store or append return value.

            Args:
                app (class): The instance of the App class "self".
            """
            data = wrapped(*args, **kwargs)
            attr = getattr(app, self.attribute)

            # tracker to indicate see if attribute has already been updated
            attr_tracker_name = f'__{self.attribute}_tracker__'
            attr_tracker = False
            try:
                attr_tracker = getattr(app, attr_tracker_name)
            except AttributeError:
                setattr(app, attr_tracker_name, False)

            # update the attribute
            if isinstance(attr, list) and isinstance(data, list) and not self.overwrite:
                attr.extend(data)
            elif isinstance(attr, list) and not self.overwrite:
                attr.append(data)
            elif attr_tracker and isinstance(data, list) and not self.overwrite:
                # convert string to list and extend with data
                d = [attr]
                d.extend(data)
                setattr(app, self.attribute, d)
            elif attr_tracker and not self.overwrite:
                # convert string to list and append with data
                d = [attr]
                d.append(data)
                setattr(app, self.attribute, d)
            else:
                setattr(app, self.attribute, data)

            # update tracker to indicate the attribute has already been set at least once
            setattr(app, attr_tracker_name, True)

            return data

        return output(instance, *args, **kwargs)



"""Decorators module for TcEx Framework"""
# flake8: noqa
from .benchmark import Benchmark
from .debug import Debug
from .fail_on_output import FailOnOutput
from .iterate_on_arg import IterateOnArg
from .on_exception import OnException
from .on_success import OnSuccess
from .output import Output
from .read_arg import ReadArg
from .write_output import WriteOutput



"""TcEx Datetime Utilities Module"""
# standard library
import calendar
import math
import re
import time
from datetime import datetime
from typing import Optional, Tuple, Union

# third-party
import parsedatetime as pdt
import pytz
from dateutil import parser
from dateutil.relativedelta import relativedelta
from tzlocal import get_localzone


class DatetimeUtils:
    """TcEx framework Datetime Utils module"""

    @staticmethod
    def _replace_timezone(dateutil_parser: object) -> object:
        """Replace the timezone on a datetime object.

        Args:
            dateutil_parser: The dateutil object.

        Returns:
            object: Update dateutils object.
        """
        try:
            # try to get the timezone from tzlocal
            tzinfo = pytz.timezone(get_localzone().zone)
        except pytz.exceptions.UnknownTimeZoneError:  # pragma: no cover
            try:
                # try to get the timezone from python's time package
                tzinfo = pytz.timezone(time.tzname[0])
            except pytz.exceptions.UnknownTimeZoneError:
                # seeing as all else has failed: use UTC as the timezone
                tzinfo = pytz.timezone('UTC')
        return tzinfo.localize(dateutil_parser)

    def any_to_datetime(self, time_input: str, tz: Optional[str] = None) -> datetime:
        """Return datetime object from multiple formats.

            Formats:

            #. Human Input (e.g 30 days ago, last friday)
            #. ISO 8601 (e.g. 2017-11-08T16:52:42Z)
            #. Loose Date format (e.g. 2017 12 25)
            #. Unix Time/Posix Time/Epoch Time (e.g. 1510686617 or 1510686617.298753)

        Args:
            time_input: The time input string (see formats above).
            tz): The time zone for the returned data.

        Returns:
            (datetime.datetime): Python datetime.datetime object.
        """
        # handle timestamp (e.g. 1510686617 or 1510686617.298753)
        dt_value: Optional[object] = self.unix_time_to_datetime(time_input, tz)

        # handle ISO or other formatted date (e.g. 2017-11-08T16:52:42Z,
        # 2017-11-08T16:52:42.400306+00:00)
        if dt_value is None:
            dt_value: Optional[object] = self.date_to_datetime(time_input, tz)

        # handle human readable relative time (e.g. 30 days ago, last friday)
        if dt_value is None:
            dt_value: Optional[object] = self.human_date_to_datetime(time_input, tz)

        # if all attempt to convert fail raise an error
        if dt_value is None:
            raise RuntimeError(f'Could not format input ({time_input}) to datetime string.')

        return dt_value

    def chunk_date_range(
        self,
        start_date: Union[int, str, datetime],
        end_date: Union[int, str, datetime],
        chunk_size: int,
        chunk_unit: Optional[str] = 'months',
        date_format: Optional[str] = None,
    ) -> Tuple[Union[datetime, str], Union[datetime, str]]:
        """Chunk a date range based on unit and size

        Args:
            start_date: Date time expression or datetime object.
            end_data: Date time expression or datetime object.
            chunk_size: Chunk size for the provided units.
            chunk_unit: A value of (years, months, days, weeks, hours, minuts, seconds)
            date_format: If None datetime object will be returned. Any other value
                must be a valid strftime format (%s for epoch seconds).

        Returns:
            Tuple[Union[datetime, str], Union[datetime, str]]: Either a datetime object
                or a string representation of the date.
        """
        # define relative delta settings
        relative_delta_settings = {chunk_unit: +chunk_size}

        # normalize inputs into datetime objects
        if isinstance(start_date, (int, str)):
            start_date = self.any_to_datetime(start_date, 'UTC')
        if isinstance(end_date, (int, str)):
            end_date = self.any_to_datetime(end_date, 'UTC')

        # set sd value for iteration
        sd = start_date
        # set ed value the the smaller of end_date or relative date
        ed = min(end_date, start_date + relativedelta(**relative_delta_settings))

        while 1:
            sdf = sd
            edf = ed
            if date_format is not None:
                # format the response data to a date formatted string
                sdf = self.format_datetime(sd.isoformat(), 'UTC', date_format)
                edf = self.format_datetime(ed.isoformat(), 'UTC', date_format)

            # yield chunked data
            yield sdf, edf

            # break iteration once chunked ed is gte to provided end_date
            if ed >= end_date:
                break

            # update sd and ed values for next iteration
            sd = ed
            ed = min(end_date, sd + relativedelta(**relative_delta_settings))

    def date_to_datetime(self, time_input: str, tz: Optional[str] = None) -> datetime:
        """Convert ISO 8601 and other date strings to datetime.datetime type.

        Args:
            time_input: The time input string (see formats above).
            tz: The time zone for the returned data.

        Returns:
            (datetime.datetime): Python datetime.datetime object.
        """
        dt = None
        try:
            # dt = parser.parse(time_input, fuzzy_with_tokens=True)[0]
            dt: object = parser.parse(time_input)
            # don't convert timezone if dt timezone already in the correct timezone
            if tz is not None and tz != dt.tzname():
                if dt.tzinfo is None:
                    dt = self._replace_timezone(dt)
                dt = dt.astimezone(pytz.timezone(tz))
        except IndexError:  # pragma: no cover
            pass
        except TypeError:  # pragma: no cover
            pass
        except ValueError:
            pass
        return dt

    def format_datetime(
        self, time_input: str, tz: Optional[str] = None, date_format: Optional[str] = None
    ) -> str:
        """Return timestamp from multiple input formats.

            Formats:

            #. Human Input (e.g 30 days ago, last friday)
            #. ISO 8601 (e.g. 2017-11-08T16:52:42Z)
            #. Loose Date format (e.g. 2017 12 25)
            #. Unix Time/Posix Time/Epoch Time (e.g. 1510686617 or 1510686617.298753)

        .. note:: To get a unix timestamp format use the strftime format **%s**. Python
                  does not natively support **%s**, however this method has support.

        Args:
            time_input: The time input string (see formats above).
            tz: The time zone for the returned data.
            date_format: The strftime format to use, ISO by default.

        Returns:
            (string): Formatted datetime string.
        """

        # handle timestamp (e.g. 1510686617 or 1510686617.298753)
        dt_value = self.any_to_datetime(time_input, tz)

        # format date
        if date_format == '%s':
            dt_value = calendar.timegm(dt_value.timetuple())
        elif date_format:
            dt_value = dt_value.strftime(date_format)
        else:
            dt_value = dt_value.isoformat()

        return dt_value

    def human_date_to_datetime(
        self, time_input: str, tz: Optional[str] = None, source_datetime: Optional[datetime] = None
    ) -> datetime:
        """Convert human readable date (e.g. 30 days ago) to datetime.datetime.

        Examples:
        * August 25th, 2008
        * 25 Aug 2008
        * Aug 25 5pm
        * 5pm August 25
        * next saturday
        * tomorrow
        * next thursday at 4pm
        * at 4pm
        * eod
        * tomorrow eod
        * eod tuesday
        * eoy
        * eom
        * in 5 minutes
        * 5 minutes from now
        * 5 hours before now
        * 2 hours before noon
        * 2 days from tomorrow

        Args:
            time_input: The time input string (see formats above).
            tz: The time zone for the returned datetime.
            source_datetime: The reference or source datetime.

        Returns:
            (datetime.datetime): Python datetime.datetime object.
        """

        c = pdt.Constants('en')
        cal = pdt.Calendar(c, version=2)
        tzinfo = None
        src_tzname = None
        if source_datetime is not None:
            tzinfo = source_datetime.tzinfo
            src_tzname = source_datetime.tzname()
        try:
            dt, status = cal.parseDT(time_input, sourceTime=source_datetime, tzinfo=tzinfo)
            if tz is not None:  # don't add tz if no tz value is passed
                if dt.tzinfo is None:
                    dt = self._replace_timezone(dt)
                # don't covert timezone if source timezone already in the correct timezone
                if tz != src_tzname:
                    dt = dt.astimezone(pytz.timezone(tz))
            if status.accuracy == 0:
                dt = None
        except TypeError:  # pragma: no cover
            dt = None

        return dt

    def timedelta(self, time_input1: str, time_input2: str) -> dict:
        """Calculate the time delta between two time expressions.

        Args:
            time_input1: The time input string (see formats above).
            time_input2: The time input string (see formats above).

        Returns:
            (dict): Dict with delta values.
        """
        time_input1: datetime = self.any_to_datetime(time_input1)
        time_input2: datetime = self.any_to_datetime(time_input2)

        diff = time_input1 - time_input2  # timedelta
        delta: object = relativedelta(time_input1, time_input2)  # relativedelta

        # totals
        total_months = (delta.years * 12) + delta.months
        total_weeks = (delta.years * 52) + (total_months * 4) + delta.weeks
        total_days = diff.days  # handles leap days
        total_hours = (total_days * 24) + delta.hours
        total_minutes = (total_hours * 60) + delta.minutes
        total_seconds = (total_minutes * 60) + delta.seconds
        total_microseconds = (total_seconds * 1000) + delta.microseconds
        return {
            'datetime_1': time_input1.isoformat(),
            'datetime_2': time_input2.isoformat(),
            'years': delta.years,
            'months': delta.months,
            'weeks': delta.weeks,
            'days': delta.days,
            'hours': delta.hours,
            'minutes': delta.minutes,
            'seconds': delta.seconds,
            'microseconds': delta.microseconds,
            'total_months': total_months,
            'total_weeks': total_weeks,
            'total_days': total_days,
            'total_hours': total_hours,
            'total_minutes': total_minutes,
            'total_seconds': total_seconds,
            'total_microseconds': total_microseconds,
        }

    @staticmethod
    def unix_time_to_datetime(time_input: str, tz: Optional[str] = None):
        """Convert timestamp into datetime.

        Convert (unix time|epoch time|posix time) in format of 1510686617
        or 1510686617.298753 to datetime.datetime type.

        .. note:: This method assumes UTC for all inputs.

        .. note:: This method only accepts a 9-10 character time_input.

        Args:
            time_input: The time input string (see formats above).
            tz: The time zone for the returned datetime (e.g. UTC).

        Returns:
            (datetime.datetime): Python datetime.datetime object.
        """
        dt = None
        if re.compile(r'^[0-9]{11,16}$').findall(str(time_input)):
            # handle timestamp with milliseconds and no "."
            time_input_length = len(str(time_input)) - 10
            dec = math.pow(10, time_input_length)
            time_input = float(time_input) / dec

        if re.compile(r'^[0-9]{9,10}(?:\.[0-9]{0,7})?$').findall(str(time_input)):
            dt = datetime.fromtimestamp(float(time_input), tz=pytz.timezone('UTC'))
            # don't covert timezone if dt timezone already in the correct timezone
            if tz is not None and tz != dt.tzname():
                dt = dt.astimezone(pytz.timezone(tz))

        return dt


# >>> from pytz import timezone
# >>> from datetime import datetime
# >>> time_input = 1229084481
# >>> dt = datetime.fromtimestamp(float(time_input), tz=timezone('UTC'))
# >>> dt.isoformat()
# '2008-12-12T12:21:21+00:00'
# >>> tz.normalize(dt).isoformat()
# '2008-12-12T06:21:21-06:00'
# >>> dt.astimezone(timezone('US/Central'))
# datetime.datetime(2008, 12, 12, 6, 21, 21,
#   tzinfo=<DstTzInfo 'US/Central' CST-1 day, 18:00:00 STD>)
# >>> dt.astimezone(timezone('US/Central')).isoformat()
# '2008-12-12T06:21:21-06:00'



"""ThreatConnect TI Document"""
from ..group import Group


class Document(Group):
    """Unique API calls for Document API Endpoints

    Args:
        tcex (TcEx): An instantiated instance of TcEx object.
        name (str, kwargs): [Required for Create] The name for this Group.
        owner (str, kwargs): The name for this Group. Default to default Org when not provided
        file_name (str, kwargs): The name for the attached file for this Group.
        malware (bool, kwargs): If true the file is considered malware.
        password (str, kwargs): If malware is true a password for the zip archive is required.
    """

    def __init__(self, tcex, **kwargs):
        """Initialize Class properties."""
        super().__init__(
            tcex, sub_type='Document', api_entity='document', api_branch='documents', **kwargs
        )

    def download(self):
        """Download the documents context.

        Returns:
            requests.Response: The response from the API call.
        """
        if not self.can_update():
            self._tcex.handle_error(910, [self.type])

        return self.tc_requests.download(self.api_type, self.api_branch, self.unique_id)

    def file_content(self, file_content, update_if_exists=True):
        """Update the file content.

        Args:
            file_content (bytes|str): The contents of the file to upload.
            update_if_exists (bool): If True the request will indicate to the API
                that the file should be updated if it exists.

        Returns:
            requests.Response
        """
        if not self.can_update():
            self._tcex.handle_error(910, [self.type])

        self._data['fileContent'] = file_content
        return self.tc_requests.upload(
            self.api_type,
            self.api_branch,
            self.unique_id,
            file_content,
            update_if_exists=update_if_exists,
        )

    def file_name(self, file_name):
        """Update the Document file name.

        Args:
            file_name (str): The filename of the document.

        Returns:
            requests.Response
        """
        if not self.can_update():
            self._tcex.handle_error(910, [self.type])

        # update data dict with new value
        self._data['fileName'] = file_name

        # build body for PUT
        request = {'fileName': file_name}
        return self.tc_requests.update(self.api_type, self.api_branch, self.unique_id, request)

    def file_size(self, file_size):
        """Set or Update the Document file size.

        Args:
            file_name (str): The filename of the document.

        Returns:
            requests.Response
        """
        if not self.can_update():
            self._tcex.handle_error(910, [self.type])

        # update data dict with new value
        self._data['fileSize'] = file_size

        # build body for PUT
        request = {'fileName': self._data.get('fileName'), 'fileSize': file_size}
        return self.tc_requests.update(self.api_type, self.api_branch, self.unique_id, request)

    def get_file_hash(self, hash_type='sha256'):
        """
        Getting the hash value of attached document
        Args:
            hash_type:

        Returns:

        """
        if not self.can_update():
            self._tcex.handle_error(910, [self.type])

        return self.tc_requests.get_file_hash(
            self.api_type, self.api_branch, self.unique_id, hash_type=hash_type
        )

    def malware(self, malware, password, file_name):
        """Update the Document file to be marked as malware.

        Args:
            malware (bool): True if the document is malware.
            password (str): The password for the zip file.
            file_name (str): The filename of the document.

        Returns:
            requests.Response
        """
        if not self.can_update():
            self._tcex.handle_error(910, [self.type])

        # update data dict with new value
        self._data['malware'] = malware
        self._data['password'] = password
        self._data['fileName'] = file_name

        # build body for PUT
        request = {'fileName': file_name, 'malware': malware, 'password': password}
        return self.tc_requests.update(self.api_type, self.api_branch, self.unique_id, request)



"""MonkeyPatch Testing Module"""
# pylint: disable=cell-var-from-loop,redefined-outer-name


def monkeypatch(target, call_target=True, tags=None):
    """Register method with the monkeypatch annotation"""

    if tags is None:
        tags = []

    def decorator(fn):
        register_monkeypatch(target=target, patch=fn, call_target=call_target, tags=tags)

        def wrapper(*args, **kwargs):
            return fn(*args, **kwargs)

        return wrapper

    return decorator


__monkeypatches = {}


def register_monkeypatch(target, patch, call_target=True, tags=None):
    """Register all methods with the monkeypatch annotation"""
    if tags is None:
        tags = []
    __monkeypatches[target.__module__ + target.__name__] = (target, patch, call_target, tags)
    # __monkeypatches.append((target, patch, call_target, tags))


def register_monkeypatches(monkeypatch, profile_data):
    """Register all methods with the monkeypatch annotation"""
    if not profile_data.get('monkeypatch'):
        return

    patches = [v for k, v in __monkeypatches.items()]

    if 'tags' in profile_data.get('monkeypatch', {}):
        include_tags = profile_data.get('monkeypatch', {}).get('tags')
        patches = filter(lambda m: set(m[3]).intersection(set(include_tags)), __monkeypatches)

    for p in patches:
        target = p[0]
        patch = p[1]
        call_target = p[2]

        def patched(*args, **kwargs):
            if call_target:
                patch(profile_data, args, kwargs)

                return target(*args, **kwargs)

            return patch(profile_data, args, kwargs)

        monkeypatch.setattr(__import__(target.__module__), target.__name__, patched)



"""ThreatConnect TI Generic Mappings Object"""
# first-party
from tcex.utils import Utils

from ..tcex_ti_tc_request import TiTcRequest


class Owner:
    """Common API calls for for Indicators/SecurityLabels/Groups and Victims"""

    def __init__(self, tcex):
        """Initialize Class properties."""
        self._tcex = tcex
        self._data = {}

        self._type = 'Owner'
        self._api_type = 'owners'
        self._api_entity = 'owner'

        self._utils = Utils()
        self._tc_requests = TiTcRequest(self._tcex)

    @property
    def type(self):
        """Return main type."""
        return self._type

    @property
    def tc_requests(self):
        """Return tc request object."""
        return self._tc_requests

    @property
    def api_type(self):
        """Return api type."""
        return self._api_type

    @property
    def api_entity(self):
        """Return api entity."""
        return self._api_entity

    @api_entity.setter
    def api_entity(self, api_entity):
        """Set the Api Entity

        Args:
            api_entity:

        Returns:

        """
        self._api_entity = api_entity

    @api_type.setter
    def api_type(self, api_type):
        """Set the Api Type

        Args:
            api_type:

        Returns:

        """
        self._api_type = api_type

    @tc_requests.setter
    def tc_requests(self, tc_requests):
        """Set the Tc Request Object

        Args:
            tc_requests:

        Returns:

        """
        self._tc_requests = tc_requests

    def single(self, unique_id):
        """Return main type."""
        return self.tc_requests.single(self.api_type, None, unique_id,)

    def many(self):
        """Get all of the owners available.

        Args:
        """
        yield from self.tc_requests.many(self.api_type, None, self.api_entity)

    def mine(self):
        """Get all of my owners available.

        Returns:

        """
        return self.tc_requests.mine()

    def members(self):
        """Get all members for my owners

        Returns:

        """
        yield from self.tc_requests.many(self.api_type, 'mine/members', 'user')

    def metrics(self):
        """Get all the metrics for owners

        Returns:

        """
        yield from self.tc_requests.many(self.api_type, 'metrics', 'ownerMetric')



"""TcEx Playbook Common module"""
# import json
# standard library
import os
import time
import uuid

from .test_case import TestCase


class TestCasePlaybookCommon(TestCase):
    """Playbook TestCase Class"""

    redis_client = None
    redis_staging_data = {
        '#App:1234:empty!Binary': '',
        '#App:1234:empty!BinaryArray': [],
        # tcex won't let these be staged
        # '#App:1234:empty!KeyValue': {},
        # '#App:1234:empty!KeyValueArray': [],
        '#App:1234:empty!String': '',
        '#App:1234:empty!StringArray': [],
        # tcex won't let these be staged
        # '#App:1234:empty!TCEntity': {},
        # '#App:1234:empty!TCEntityArray': [],
        '#App:1234:non-ascii!String': 'ドメイン.テスト',
        # staging the data is not required
        # '#App:1234:null!Binary': None,
        # '#App:1234:null!BinaryArray': None,
        # '#App:1234:null!KeyValue': None,
        # '#App:1234:null!KeyValueArray': None,
        # '#App:1234:null!String': None,
        # '#App:1234:null!StringArray': None,
        # '#App:1234:null!TCEntity': None,
        # '#App:1234:null!TCEntityArray': None,
    }

    def setup_method(self):
        """Run before each test method runs."""
        super().setup_method()
        # APP-78 - The unique redis data that will be regenerated each test method.
        redis_unique_staged_data = {
            '#App:1234:epoch!String': str(int(time.time())),
            '#App:1234:uuid!String': str(uuid.uuid4()),
        }
        self.redis_staging_data.update(redis_unique_staged_data)

    def clear_context(self, context):
        """Delete data in redis"""
        keys = self.redis_client.hkeys(context)
        if keys:
            return self.redis_client.hdel(context, *keys)
        return 0

    @property
    def default_args(self):
        """Return App default args."""
        args = super().default_args.copy()
        args.update(
            {
                'tc_playbook_db_context': self.context,
                'tc_playbook_db_path': os.getenv('TC_PLAYBOOK_DB_PATH', 'localhost'),
                'tc_playbook_db_port': os.getenv('TC_PLAYBOOK_DB_PORT', '6379'),
                'tc_playbook_db_type': os.getenv('TC_PLAYBOOK_DB_TYPE', 'Redis'),
                'tc_playbook_out_variables': '',
            }
        )
        return args

    def run(self):
        """Implement in Child Class"""
        raise NotImplementedError('Child class must implement this method.')

    def stage_data(self, staged_data):
        """Stage the data in the profile."""
        for key, value in list(staged_data.get('redis', {}).items()):
            self.stager.redis.stage(key, value)

    def teardown_method(self):
        """Run after each test method runs."""
        if self.enable_update_profile:
            self.log.title('Update Outputs')
            self.profile.update_outputs()

        # clear context tracker
        if self.profile:  # doesn't exist for API services
            self.profile._context_tracker = []

        # run test_case teardown_method
        super().teardown_method()



"""ThreatConnect TQL"""
# standard library
from enum import Enum

from .filter import Filter


class TQL:
    """ThreatConnect TQL"""

    class Operator(Enum):
        """Available TQL Operators"""

        EQ = '='
        NE = '!='
        GT = '>'
        LT = '<'
        LEQ = '<='
        GEQ = '>='
        NOT_IN = 'NOT IN'
        IN = 'IN'
        NOT_LIKE = 'NOT LIKE'
        LIKE = 'LIKE'
        NOT_CONTAINS = 'NOT CONTAINS'
        CONTAINS = 'CONTAINS'
        NOT_STARTS_WITH = 'NOT STARTSWITH'
        STARTS_WITH = 'STARTSWITH'
        NOT_ENDS_WITH = 'NOT ENDSWITH'
        ENDS_WITH = 'ENDSWITH'

    class Type(Enum):
        """Enum representing available value types"""

        STRING = 'String'
        INTEGER = 'Integer'
        BOOLEAN = 'Boolean'
        SUB_QUERY = 'Sub Query'

    def __init__(self):
        """Initialize Class Properties"""
        self._filters = []
        self.raw_tql = None

    @property
    def as_str(self):
        """Convert the TQL obj to a string"""
        filters = []
        for tql_filter in self.filters:
            value = tql_filter.get('value')
            # keyword = tql_filter.get('keyword')
            if isinstance(value, Filter):
                filters.append(f"{tql_filter.get('keyword')}({value._tql.as_str})")
            # elif keyword.startswith('has'):
            #     values = value
            #     if not isinstance(value, list):
            #         values = [value]
            #     if tql_filter.get('type') == self.Type.STRING:
            #         values = [f'"{value}"' for value in values]
            #     # value = f"({','.join(values)})"
            #     value = f"({','.join([str(v) for v in values])})"
            #     if tql_filter.get('type') == self.Type.STRING:
            #         value = f'"{value}"'
            #     filters.append(
            #         f"{tql_filter.get('keyword')} {tql_filter.get('operator').name} {value}"
            #     )
            else:
                if tql_filter.get('type') == self.Type.STRING:
                    value = f'"{value}"'
                filters.append(
                    f"{tql_filter.get('keyword')} {tql_filter.get('operator').name} {value}"
                )

        return ' and '.join(filters)

    @property
    def filters(self):
        """Return the filters"""
        return self._filters

    @filters.setter
    def filters(self, filters):
        """Set the filters"""
        self._filters = filters

    def add_filter(self, keyword, operator, value, type_=Type.STRING):
        """Add a filter to the current obj

        Args:
            keyword (str): the field to search on
            operator (str): the operator to use
            value (str): the value to compare
            type_ (Type): How to treat the value (defaults to String)
        """
        self.filters.append(
            {'keyword': keyword, 'operator': operator, 'value': value, 'type': type_}
        )

    def set_raw_tql(self, tql):
        """Set a raw TQL filter"""
        self.raw_tql = tql



"""TcEx Framework Service Trigger Common module."""
# standard library
import json
import os
import threading
import traceback
from typing import Callable, Optional, Union

from .common_service import CommonService


class CommonServiceTrigger(CommonService):
    """TcEx Framework Service Trigger Common module.

    Shared service logic between the supported service types:
    * Custom Trigger Service
    * Webhook Trigger Service
    """

    def __init__(self, tcex: object):
        """Initialize the Class properties.

        Args:
            tcex: Instance of TcEx.
        """
        super().__init__(tcex)

        # properties
        self._metrics = {'Active Playbooks': 0, 'Errors': 0, 'Hits': 0, 'Misses': 0}
        self.configs = {}
        self.config_thread = None

        # config callbacks
        self.create_config_callback = None
        self.delete_config_callback = None

    def _tcex_testing(self, session_id: str, trigger_id: int) -> None:
        """Write data required for testing framework to Redis.

        Args:
            session_id: The context/session id value for the current operation.
            trigger_id: The trigger ID for the current playbook.
        """
        if self.args.tcex_testing_context is not None:
            _context_tracker: str = (
                self.redis_client.hget(self.args.tcex_testing_context, '_context_tracker') or '[]'
            )
            _context_tracker = json.loads(_context_tracker)
            _context_tracker.append(session_id)
            self.redis_client.hset(
                self.args.tcex_testing_context, '_context_tracker', json.dumps(_context_tracker),
            )
            self.redis_client.hset(session_id, '_trigger_id', trigger_id)

            # log
            self.log.info(
                'feature=service, event=testing-context-tracker, '
                f'context={session_id}, trigger-id={trigger_id}'
            )

    def _tcex_testing_fired_events(self, session_id: str, fired: bool) -> None:
        """Write fired event data to KV Store to be used in test validation.

        Args:
            session_id: The context/session id value for the current operation.
            fired: The value to increment the count by.
        """
        if self.args.tcex_testing_context is not None:
            self.redis_client.hset(
                session_id, '#Trigger:9876:_fired!String', json.dumps(str(fired).lower())
            )

    @property
    def command_map(self) -> dict:
        """Return the command map for the current Service type."""
        command_map = super().command_map
        command_map.update(
            {
                'createconfig': self.process_create_config_command,
                'deleteconfig': self.process_delete_config_command,
            }
        )
        return command_map

    def create_config(self, trigger_id: int, message: str, status: bool) -> None:
        """Add config item to service config object.

        Args:
            trigger_id: The trigger ID for the current config.
            message: A simple message for the action.
            status: The passed/fail status for the App handling of config.
            logfile: The CreateConfig logfile to return in response ack.
        """
        try:
            if status is not True and self.configs.get(str(trigger_id)) is not None:
                # add config to configs
                del self.configs[str(trigger_id)]

            # send ack response
            self.message_broker.publish(
                json.dumps(
                    {
                        'command': 'Acknowledged',
                        'logFile': os.path.join(
                            os.path.basename(os.path.dirname(self.trigger_logfile)),
                            os.path.basename(self.trigger_logfile),
                        ),
                        'message': message,
                        'status': 'Success' if status is True else 'Failed',
                        'type': 'CreateConfig',
                        'triggerId': trigger_id,
                    }
                ),
                self.args.tc_svc_client_topic,
            )
        except Exception as e:
            self.log.error(
                'feature=service, event=create-config-callback-exception, '
                f'trigger-id={trigger_id}, error="""{e}"""'
            )
            self.log.trace(traceback.format_exc())

    def delete_config(self, trigger_id: int, message: str, status: str) -> None:
        """Delete config item from config object.

        Args:
            trigger_id: The trigger ID for the current config.
            message: A simple message for the action.
            status: The passed/fail status for the App handling of config.
        """
        try:
            # always delete config from configs dict, even when status is False
            del self.configs[trigger_id]

            # send ack response
            self.message_broker.publish(
                json.dumps(
                    {
                        'command': 'Acknowledged',
                        'message': message,
                        'status': 'Success' if status is True else 'Failed',
                        'type': 'DeleteConfig',
                        'triggerId': trigger_id,
                    }
                ),
                self.args.tc_svc_client_topic,
            )
        except Exception as e:
            self.log.error(
                'feature=service, event=delete-config-callback-exception, '
                f'trigger-id={trigger_id}, error="""{e}"""'
            )
            # self.log.trace(traceback.format_exc())

    def fire_event(self, callback: Callable[[], bool], **kwargs) -> None:
        """Trigger a FireEvent command.

        Args:
            callback: The trigger method in the App to call.
            trigger_ids: A list of trigger ids to trigger.
        """
        if not callable(callback):
            raise RuntimeError('Callback method (callback) is not a callable.')

        # get developer passed trigger_ids
        trigger_ids: Optional[list] = kwargs.pop('trigger_ids', None)

        for trigger_id, config in list(self.configs.items()):
            if trigger_ids is not None and trigger_id not in trigger_ids:
                # skip config that don't match developer provided trigger ids
                continue

            try:
                # get a session_id specifically for this thread
                session_id: str = self.create_session_id()

                # only required for testing in tcex framework
                self._tcex_testing(session_id, trigger_id)

                # get instance of playbook specifically for this thread
                outputs: Union[list, str] = config.get('tc_playbook_out_variables') or []
                if isinstance(outputs, str):
                    outputs = outputs.split(',')

                # get an instance of PB module with current session_id to pass to callback
                playbook: object = self.tcex.pb(context=session_id, output_variables=outputs)

                self.log.info(f'feature=trigger-service, event=fire-event, trigger-id={session_id}')

                # current thread has session_id as name
                self.service_thread(
                    name=session_id,
                    target=self.fire_event_trigger,
                    args=(callback, playbook, session_id, trigger_id, config,),
                    kwargs=kwargs,
                    session_id=session_id,
                    trigger_id=trigger_id,
                )
            except Exception:
                self.log.trace(traceback.format_exc())

    def fire_event_publish(
        self, trigger_id: int, session_id: str, request_key: Optional[str] = None
    ) -> None:
        """Send FireEvent command.

        Args:
            trigger_id: The ID of the trigger.
            session_id: The generated session for this fired event.
            request_key: The request key for this response.
        """
        msg = {
            'command': 'FireEvent',
            'triggerId': trigger_id,  # reference to single playbook
            'sessionId': session_id,  # session for the playbook execution
        }
        if request_key is not None:
            msg['requestKey'] = request_key  # reference for a specific playbook execution
        self.log.info(f'feature=service, event=fire-event, msg={msg}')

        # publish FireEvent command to client topic
        self.message_broker.publish(json.dumps(msg), self.args.tc_svc_client_topic)

    def fire_event_trigger(
        self,
        callback: Callable[[], bool],
        playbook: object,
        session_id: str,
        trigger_id: int,
        config: dict,
        **kwargs: str,
    ) -> None:
        """Fire event for trigger.

        Args:
            callback: The App callback method for firing an event.
            playbook: A configure playbook instance for using to interact with KvStore.
            session_id: The current session Id.
            trigger_id: The current trigger Id.
            config: A dict containing the configuration information.
        """
        self._create_logging_handler()
        self.log.info('feature=trigger-service, event=fire-event-trigger')

        try:
            if callback(playbook, trigger_id, config, **kwargs):
                self.increment_metric('Hits')
                self.fire_event_publish(trigger_id, session_id)

                # capture fired status for testing framework
                self._tcex_testing_fired_events(session_id, True)
            else:
                self.increment_metric('Misses')
                self.log.info(
                    'feature=trigger-service, event=fire-event-callback-miss, '
                    f'trigger-id={trigger_id}'
                )

                # capture fired status for testing framework
                self._tcex_testing_fired_events(session_id, False)
        except Exception as e:
            self.increment_metric('Errors')
            self.log.error(
                f'feature=trigger-service, event=fire-event-callback-exception, error="""{e}"""'
            )
            self.log.trace(traceback.format_exc())
        finally:
            self.logger.remove_handler_by_name(self.thread_name)

    def process_create_config_command(self, message: dict) -> None:
        """Process the CreateConfig command.

        .. code-block:: python
            :linenos:
            :lineno-start: 1

            {
                "appId": 387,
                "command": "CreateConfig",
                "triggerId": 1,
                "config": {
                    "password": "test-pass",
                    "username": "test-user",
                    "cc_action": "pass",
                    "tc_playbook_out_variables": "#Trigger:1:testing.body!String"
                },
                "apiToken": "SVC:8:QQuyIp:1596817138182:387:+9vBOAT8Y56caHRcjLa4IwAqABoatsYOU ... ",
                "expireSeconds": 1596817138
            }

        Args:
            message: The message payload from the server topic.
        """
        config: dict = message.get('config')
        status = True
        trigger_id = int(message.get('triggerId'))

        # create trigger id logging filehandler
        self.logger.add_thread_file_handler(
            backup_count=1,  # required for logs to be rotated
            name=self.thread_name,
            filename=self.trigger_logfile,
            level=self.args.tc_log_level,
            max_bytes=1048576,  # 1Mb
            path=self.args.tc_log_path,
            handler_key=trigger_id,
            thread_key='trigger_id',
        )

        self.log.info(
            f'feature=service, event=create-config, trigger_id={trigger_id}, config={config}'
        )

        # register config apiToken
        self.token.register_token(trigger_id, message.get('apiToken'), message.get('expireSeconds'))

        # temporarily add config, will be removed if callback fails
        self.configs[trigger_id] = config

        msg = 'Create Config'
        if callable(self.create_config_callback):

            kwargs = {}
            if self.ij.runtime_level.lower() == 'webhooktriggerservice':
                # only webhook triggers get and require the PB url
                kwargs['url'] = message.get('url')

            try:
                # call callback for create config and handle exceptions to protect thread
                # pylint: disable=not-callable
                response: Optional[dict] = self.create_config_callback(trigger_id, config, **kwargs)
                if isinstance(response, dict):
                    status = response.get('status', False)
                    msg = response.get('msg')

                # if callback does not return a boolean value assume it worked
                if not isinstance(status, bool):
                    status = True
            except Exception as e:
                status = False
                msg = str(e)
                self.log.error(
                    f'feature=service, event=create-config-callback-exception, error="""{e}"""'
                )
                self.log.error(message)
                self.log.trace(traceback.format_exc())

        # create config after callback to report status and message
        self.create_config(trigger_id, msg, status)

    def process_delete_config_command(self, message: dict) -> None:
        """Process the DeleteConfig command.

        .. code-block:: python
            :linenos:
            :lineno-start: 1

            {
                "appId": 387,
                "command": "DeleteConfig",
                "triggerId": 1
            }

        Args:
            message: The message payload from the server topic.
        """
        status = True
        trigger_id = int(message.get('triggerId'))
        self.log.info(f'feature=service, event=delete-config, trigger_id={trigger_id}')

        # unregister config apiToken
        self.token.unregister_token(trigger_id)

        msg = 'Delete Config'
        if callable(self.delete_config_callback):
            try:
                # call callback for delete config and handle exceptions to protect thread
                # pylint: disable=not-callable
                status: Optional[bool] = self.delete_config_callback(trigger_id)

                # if callback does not return a boolean value assume it worked
                if not isinstance(status, bool):
                    status = True
            except Exception as e:
                self.log.error(
                    f'feature=service, event=delete-config-callback-exception, error="""{e}"""'
                )
                self.log.trace(traceback.format_exc())
                status = False

        # delete config
        self.delete_config(trigger_id, msg, status)

        # remove temporary logging file handler
        self.logger.remove_handler_by_name(self.thread_name)

    @property
    def trigger_logfile(self) -> str:
        """Return the logfile name based on date and thread name."""
        return f'''trigger-id-{self.thread_trigger_id}.log'''

    @property
    def thread_trigger_id(self) -> Optional[str]:
        """Return the current thread trigger id."""
        trigger_id = None
        if hasattr(threading.current_thread(), 'trigger_id'):
            trigger_id = threading.current_thread().trigger_id
        return trigger_id



"""ThreatConnect TI Event"""
from ..group import Group


class Event(Group):
    """Unique API calls for Event API Endpoints

    Valid status:
    + Escalated
    + False Positive
    + Needs Review
    + No Further Action

    Args:
        tcex (TcEx): An instantiated instance of TcEx object.
        event_date (str, kwargs): The event "event date" datetime expression for this Group.
        name (str, kwargs): [Required for Create] The name for this Group.
        owner (str, kwargs): The name for this Group. Default to default Org when not provided
        status (str, kwargs): The status for this Group.
    """

    def __init__(self, tcex, **kwargs):
        """Initialize Class Properties."""
        super().__init__(tcex, sub_type='Event', api_entity='event', api_branch='events', **kwargs)

    def event_date(self, event_date):
        """Update the event date for the Event.

        Args:
            event_date (str): The event datetime expression for this Group.

        Returns:
            requests.Response: The response from the API call.
        """
        if not self.can_update():
            self._tcex.handle_error(910, [self.type])

        event_date = self._utils.datetime.format_datetime(
            event_date, date_format='%Y-%m-%dT%H:%M:%SZ'
        )
        self._data['eventDate'] = event_date
        request = {'eventDate': event_date}
        return self.tc_requests.update(self.api_type, self.api_branch, self.unique_id, request)

    def status(self, status):
        """Update the event date for the Event.

        Valid status:
        + Escalated
        + False Positive
        + Needs Review
        + No Further Action

        Args:
            status (str, kwargs): The status for this Group.

        Returns:
            requests.Response: The response from the API call.
        """
        if not self.can_update():
            self._tcex.handle_error(910, [self.type])

        self._data['status'] = status
        request = {'status': status}
        return self.tc_requests.update(self.api_type, self.api_branch, self.unique_id, request)



"""TcEx Utilities Module"""
# standard library
import ipaddress
import os
import random
import re
import string
import tempfile
import uuid
from typing import Any, List, Optional, Union
from urllib.parse import urlsplit

# third-party
import jmespath
import pyaes

from .date_utils import DatetimeUtils
from .mitre_attack_utils import MitreAttackUtils


class Utils:
    """TcEx framework Utils Class

    Args:
        temp_path (Optional[str] = None): The path to write temp files.
    """

    def __init__(self, temp_path: Optional[str] = None):
        """Initialize the Class properties."""
        self.temp_path = temp_path or tempfile.gettempdir() or '/tmp'  # nosec

        # properties
        self._camel_pattern = re.compile(r'(?<!^)(?=[A-Z])')
        self._inflect = None
        self.variable_match = re.compile(fr'^{self.variable_pattern}$')
        self.variable_parse = re.compile(self.variable_pattern)

    def camel_to_snake(self, camel_string: str) -> str:
        """Return snake case string from a camel case string.

        Args:
            camel_string (str): The camel case input string.

        Returns:
            str: The snake case representation of input string.
        """
        return self._camel_pattern.sub('_', camel_string).lower()

    def camel_to_space(self, camel_string: str) -> str:
        """Return space case string from a camel case string.

        Args:
            camel_string (str): The camel case input string.

        Returns:
            str: The space representation of input string.
        """
        return self._camel_pattern.sub(' ', camel_string).lower()

    @property
    def datetime(self) -> object:
        """Return an instance of DatetimeUtils."""
        return DatetimeUtils()

    @property
    def mitre_attack(self) -> object:
        """Return an instance of MitreAttackUtils."""
        return MitreAttackUtils()

    @staticmethod
    def decrypt_aes_cbc(
        key: bytes, ciphertext: Union[bytes, str], iv: Optional[bytes] = None
    ) -> bytes:
        """Return AES CBC decrypted string.

        Args:
            key (bytes): The encryption key.
            ciphertext (Optional[bytes] = None): The ciphertext to decrypt.
            iv (bytes, optional): The CBC initial vector.

        Returns:
            bytes: The encoded string.
        """
        iv = iv or b'\0' * 16

        # ensure key is bytes
        if isinstance(key, str):
            key = key.encode()

        # ensure iv is bytes
        if isinstance(iv, str):
            iv = iv.encode()

        aes_cbc_decrypt = pyaes.Decrypter(pyaes.AESModeOfOperationCBC(key, iv=iv))
        decrypted = aes_cbc_decrypt.feed(ciphertext)
        decrypted += aes_cbc_decrypt.feed()
        return decrypted

    @staticmethod
    def encrypt_aes_cbc(
        key: bytes, plaintext: Union[bytes, str], iv: Optional[bytes] = None
    ) -> bytes:
        """Return AES CBC encrypted string.

        Args:
            key (bytes): The encryption key.
            plaintext (Union[bytes, str]): The text to encrypt.
            iv (Optional[bytes] = None): The CBC initial vector.

        Returns:
            bytes: The encoded string.
        """
        iv = iv or b'\0' * 16

        # ensure key is bytes
        if isinstance(key, str):
            key = key.encode()

        # ensure plaintext is bytes
        if isinstance(plaintext, str):
            plaintext = plaintext.encode()

        # ensure iv is bytes
        if isinstance(iv, str):
            iv = iv.encode()

        aes_cbc_encrypt = pyaes.Encrypter(pyaes.AESModeOfOperationCBC(key, iv=iv))
        encrypted = aes_cbc_encrypt.feed(plaintext)
        encrypted += aes_cbc_encrypt.feed()
        return encrypted

    @staticmethod
    def flatten_list(lst: List[Any]) -> List[Any]:
        """Flatten a list

        Will work for lists of lists to arbitrary depth
        and for lists with a mix of lists and single values

        Args:
            lst (list): The list to flatten

        Returns:
            list: the flattened list
        """
        flat_list = []
        for sublist in lst:
            if isinstance(sublist, list):
                for item in Utils.flatten_list(sublist):
                    flat_list.append(item)
            else:
                flat_list.append(sublist)

        return flat_list

    @property
    def inflect(self) -> object:
        """Return instance of inflect."""
        if self._inflect is None:
            # third-party
            import inflect

            self._inflect = inflect.engine()
        return self._inflect

    @staticmethod
    def is_cidr(possible_cidr_range: str) -> bool:
        """Return True if the provided value is a valid CIDR block.

        Args:
            possible_cidr_range (str): The cidr value to validate.

        Returns:
            bool: True if input is a valid CIDR.
        """
        try:
            ipaddress.ip_address(possible_cidr_range)
        except ValueError:
            try:
                ipaddress.ip_interface(possible_cidr_range)
            except Exception:
                return False
            else:
                return True
        else:
            return False

    @staticmethod
    def is_ip(possible_ip: str) -> bool:
        """Return True if the provided value is a valid IP address.

        Args:
            possible_ip (str): The IP value to validate.

        Returns:
            bool: True if input is a valid IP.
        """
        try:
            ipaddress.ip_address(possible_ip)
        except ValueError:
            return False
        else:
            return True

    @staticmethod
    def printable_cred(
        cred: str,
        visible: Optional[int] = 1,
        mask_char: Optional[str] = '*',
        mask_char_count: Optional[int] = 4,
    ) -> str:
        """Return a printable (masked) version of the provided credential.

        Args:
            cred: The cred to print.
            visible: The number of characters at the beginning and ending of the cred to not mask.
            mask_char: The character to use in the mask.
            mask_char_count: How many mask character to insert (obscure cred length).

        Returns:
            str: The reformatted token.
        """
        mask_char = mask_char or '*'
        if cred is not None and len(cred) >= visible * 2:
            cred = f'{cred[:visible]}{mask_char * mask_char_count}{cred[-visible:]}'
        return cred

    @staticmethod
    def random_string(string_length: Optional[int] = 10) -> str:
        """Generate a random string of fixed length

        Args:
            string_length (Optional[int] = 10): The length of the string.

        Returns:
            str: A random string
        """
        return ''.join(random.choice(string.ascii_letters) for _ in range(string_length))  # nosec

    def requests_to_curl(
        self,
        request: object,
        mask_headers: Optional[bool] = True,
        mask_patterns: List[str] = None,
        **kwargs: Union[bool, dict],
    ) -> str:
        """Return converted PreparedRequest to a curl command.

        Args:
            request (object): The response.request object.
            mask_headers (Optional[bool] = True): If True then
                values for certain header key will be masked.
            mask_patterns (list[str] = None): A list of patterns
                if found in headers the value will be masked.
            body_limit (int, kwargs): The size limit for the body value.
            mask_body (bool, kwargs): If True the body will be masked.
            proxies (dict, kwargs): A dict containing the proxy configuration.
            verify (bool, kwargs): If False the curl command will include --insecure flag.
            write_file (bool, kwargs): If True and the body is
                binary it will be written as a temp file.

        Returns:
            str: The curl command.
        """
        body_limit: int = kwargs.get('body_limit', 100)
        proxies: dict = kwargs.get('proxies', {})
        verify: bool = kwargs.get('verify', True)
        write_file: bool = kwargs.get('write_file', False)

        # APP-79 - adding the ability to log request as curl commands
        cmd = ['curl', '-X', request.method]

        # add headers to curl command
        for k, v in sorted(list(dict(request.headers).items())):
            if mask_headers is True:
                patterns = [
                    'authorization',
                    'cookie',
                    'password',
                    'secret',
                    'session',
                    'username',
                    'token',
                ]
                if isinstance(mask_patterns, list):
                    # add user defined mask patterns
                    patterns.extend(mask_patterns)

                for p in patterns:
                    if re.match(rf'.*{p}.*', k, re.IGNORECASE):
                        v: str = self.printable_cred(v)

                # using gzip in Accept-Encoding with CURL on the CLI produces
                # the warning "Binary output can mess up your terminal."
                if k.lower() == 'accept-encoding':
                    encodings = [e.strip() for e in v.split(',')]
                    for encoding in list(encodings):
                        if encoding in ['gzip']:
                            encodings.remove(encoding)
                    v: str = ', '.join(encodings)

            cmd.append(f"-H '{k}: {v}'")

        if request.body:
            # add body to the curl command
            body = request.body
            try:
                if isinstance(body, bytes):
                    body = body.decode('utf-8')

                if kwargs.get('mask_body', False):
                    # mask_body
                    body = self.printable_cred(body)
                else:
                    # truncate body
                    body = self.truncate_string(
                        t_string=body, length=body_limit, append_chars='...'
                    )
                body_data = f'-d "{body}"'
            except Exception:
                # set static filename so that when running a large job App thousands of files do
                # no get created.
                body_data = '--data-binary @/tmp/body-file'
                if write_file is True:
                    temp_file: str = self.write_temp_binary_file(content=body, filename='curl-body')
                    body_data = f'--data-binary @{temp_file}'
            cmd.append(body_data)

        if proxies is not None and proxies.get('https'):
            # parse formatted string {'https': 'bob:pass@https://localhost:4242'}
            proxy_url: Optional[str] = proxies.get('https')
            proxy_data = urlsplit(proxy_url)

            # auth
            if proxy_data.username:
                cmd.extend(['--proxy-user', f'{proxy_data.username}:xxxxx'])

            # server
            proxy_server = proxy_data.hostname
            if proxy_data.port:
                proxy_server = f'{proxy_data.hostname}:{proxy_data.port}'
            cmd.extend(['--proxy', proxy_server])

        if not verify:
            # add insecure flag to curl command
            cmd.append('--insecure')

        # add url to curl command
        cmd.append(request.url)

        return ' '.join(cmd)

    @staticmethod
    def snake_to_camel(snake_string: str) -> str:
        """Convert snake_case to camelCase

        Args:
            snake_string (str): The snake case input string.

        Returns:
            str: The camel case string.
        """
        components = snake_string.split('_')
        return components[0] + ''.join(x.title() for x in components[1:])

    @staticmethod
    def standardize_asn(asn: str) -> str:
        """Return the ASN formatted for ThreatConnect.

        Args:
            asn (str): The asn value to standardize.

        Returns:
            str: The standardized asn value.
        """
        numbers = re.findall('[0-9]+', asn)
        if len(numbers) == 1:
            asn = f'ASN{numbers[0]}'
        return asn

    @staticmethod
    def to_bool(value: Union[bool, int, str]) -> bool:
        """Convert value to bool.

        Args:
            value (Union[bool, int, str]): The value to convert to boolean.

        Returns:
            bool: The boolean value
        """
        return str(value).lower() in ['1', 't', 'true', 'y', 'yes']

    @staticmethod
    def truncate_string(
        t_string: str,
        length: int,
        append_chars: Optional[str] = '',
        spaces: Optional[bool] = False,
    ) -> str:
        """Truncate a string to a given length.

        Args:
            t_string (str): The input string to truncate.
            length (int): The length of the truncated string.
            append_chars (Optional[str] = ''): Any character that should be appended to the
                string. Typically used for ellipsis (e.g. ...).
            spaces (Optional[bool] = False): If True truncation will be done at the
                nearest space before the truncation length to avoid chopping words.

        Returns:
            str: The truncated string.
        """
        if t_string is None:
            t_string = ''

        if length is None:
            length = len(t_string)

        if len(t_string) <= length:
            return t_string

        # set sane default for append_chars
        append_chars = str(append_chars or '')

        # ensure append_chars is not longer than length
        if len(append_chars) > length:  # pragma: no cover
            raise RuntimeError('Append chars cannot exceed the truncation length.')

        output = t_string[0 : length - len(append_chars)]  # noqa: E203
        if spaces is True:
            if not output.endswith(' '):
                # split output on spaces and drop last item to terminate string on word
                output = ' '.join(output.split(' ')[:-1])

        return f'{output.rstrip()}{append_chars}'

    def variable_method_name(self, variable: str) -> str:
        """Convert variable name to a valid method name.

        #App:9876:string.operation!String -> string_operation_string

        Args:
            variable (string): The variable name to convert.

        Returns:
            (str): Method name
        """
        method_name = None
        if variable is not None:
            variable = variable.strip()
            if re.match(self.variable_match, variable):
                var = re.search(self.variable_parse, variable)
                variable_name = var.group(3).replace('.', '_').lower()
                variable_type = var.group(4).lower()
                method_name = f'{variable_name}_{variable_type}'
        return method_name

    @property
    def variable_pattern(self) -> str:
        """Regex pattern to match and parse a playbook variable."""
        return (
            r'#([A-Za-z]+)'  # match literal (#App) at beginning of String
            r':([\d]+)'  # app id (:7979)
            r':([A-Za-z0-9_\.\-\[\]]+)'  # variable name (:variable_name)
            r'!(StringArray|BinaryArray|KeyValueArray'  # variable type (array)
            r'|TCEntityArray|TCEnhancedEntityArray'  # variable type (array)
            r'|String|Binary|KeyValue|TCEntity|TCEnhancedEntity'  # variable type
            r'|(?:(?!String)(?!Binary)(?!KeyValue)'  # non matching for custom
            r'(?!TCEntity)(?!TCEnhancedEntity)'  # non matching for custom
            r'[A-Za-z0-9_-]+))'  # variable type (custom)
        )

    def write_temp_binary_file(self, content: bytes, filename: Optional[str] = None) -> str:
        """Write content to a temporary file.

        Args:
            content (bytes): The file content.
            filename (Optional[str] = None): The filename to use when writing the file.

        Returns:
            str: Fully qualified path name for the file.
        """
        return self.write_temp_file(content, filename, 'wb')

    def write_temp_file(
        self, content: Union[bytes, str], filename: Optional[str] = None, mode: Optional[str] = 'w'
    ) -> str:
        """Write content to a temporary file.

        If passing binary data the mode needs to be set to 'wb'.

        Args:
            content (Union[bytes, str]): The file content.
            filename (Optional[str] = None): The filename to use when writing the file.
            mode (Optional[str] = 'w'): The write mode ('w' or 'wb').

        Returns:
            str: Fully qualified path name for the file.
        """
        if filename is None:
            filename = str(uuid.uuid4())
        fqpn = os.path.join(self.temp_path, filename)
        os.makedirs(os.path.dirname(fqpn), exist_ok=True)
        with open(fqpn, mode) as fh:
            fh.write(content)
        return fqpn

    def mapper(self, data: Union[list, dict], mapping: dict):
        """Yield something ..."""
        # TODO - @bpurdy - update docstring
        if isinstance(data, dict):
            data = [data]
        try:
            for d in data:
                mapped_obj = mapping.copy()
                for key, value in mapping.items():
                    if isinstance(value, list):
                        new_list = []
                        for item in value:
                            if isinstance(item, dict):
                                new_list.append(list(self.mapper(d, item))[0])
                            else:
                                if not item.startswith('@'):
                                    new_list.append(item)
                                else:
                                    new_list.append(
                                        jmespath.search(f'{item}', jmespath.search('@', d))
                                    )

                        mapped_obj[key] = new_list
                    elif isinstance(value, dict):
                        mapped_obj[key] = list(self.mapper(d, mapped_obj[key]))[0]
                    else:
                        if not value.startswith('@'):
                            mapped_obj[key] = value
                        else:
                            mapped_obj[key] = jmespath.search(f'{value}', jmespath.search('@', d))
                yield mapped_obj
        except Exception:  # nosec
            pass



"""ThreatConnect Requests Session"""
# standard library
import logging
import time
from typing import Callable, Optional

# third-party
import urllib3
from requests import Response, Session, adapters, exceptions
from requests.adapters import DEFAULT_POOLBLOCK, DEFAULT_POOLSIZE, DEFAULT_RETRIES
from urllib3.util.retry import Retry

from ..utils import Utils
from .rate_limit_handler import RateLimitHandler

# disable ssl warning message
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


def default_too_many_requests_handler(response: Response) -> float:
    """Implement 429 response handling that uses the Retry-After header.

    Will return the value in Retry-After.  See: https://tools.ietf.org/html/rfc6585#page-3.

    Assumptions:
        - Response has a Retry-After header.

    Args:
        response: The 429 response.

    Returns:
        The number of seconds to wait before sending the next request, from the Retry-After header.
    """
    utils = Utils()
    retry_after = response.headers.get('Retry-After', 0)
    try:
        seconds = (
            float(utils.datetime.format_datetime(float(retry_after), date_format='%s'))
            - time.time()
        )
    except RuntimeError:
        # retry_after must be in seconds
        seconds = retry_after
    return float(seconds)


class CustomAdapter(adapters.HTTPAdapter):
    """Custom Adapter to properly handle retries."""

    def __init__(
        self,
        rate_limit_handler: Optional[RateLimitHandler] = None,
        pool_connections=DEFAULT_POOLSIZE,
        pool_maxsize=DEFAULT_POOLSIZE,
        max_retries=DEFAULT_RETRIES,
        pool_block=DEFAULT_POOLBLOCK,
    ):
        """Initialize CustomAdapter.

        Args:
            rate_limit_handler: RateLimitHandler responsible for throttling.
            pool_connections: passed to super
            pool_maxsize: passed to super
            max_retries: passed to super
            pool_block: passed to super
        """
        super().__init__(pool_connections, pool_maxsize, max_retries, pool_block)
        self._rate_limit_handler = rate_limit_handler

    def send(self, request, stream=False, timeout=None, verify=True, cert=None, proxies=None):
        """Send PreparedRequest object. Returns Response object."""
        if self.rate_limit_handler:
            self.rate_limit_handler.pre_send(request)

        try:
            response = super().send(request, stream, timeout, verify, cert, proxies)
        except exceptions.RetryError:
            # store current retries configuration
            max_retries = self.max_retries

            # temporarily disable retries and make one last request
            self.max_retries = Retry(0, read=False)

            # make request with max_retries turned off
            response = super().send(request, stream, timeout, verify, cert, proxies)

            # reset retries configuration
            self.max_retries = max_retries

        if self.rate_limit_handler:
            self.rate_limit_handler.post_send(response)

        return response

    @property
    def rate_limit_handler(self) -> RateLimitHandler:
        """Get the RateLimitHandler."""
        return self._rate_limit_handler

    @rate_limit_handler.setter
    def rate_limit_handler(self, rate_limit_handler: RateLimitHandler) -> None:
        """Set the RateLimitHandler."""
        self._rate_limit_handler = rate_limit_handler


class ExternalSession(Session):
    """ThreatConnect REST API Requests Session for external requests

    Args:
        base_url (Optional[str] = None): The base URL for all requests.
        logger (Optional[object] = None): An instance of Logger.
    """

    __attrs__ = [
        'adapters',
        'auth',
        'cert',
        'cookies',
        'headers',
        'hooks',
        'max_redirects',
        'proxies',
        'params',
        'stream',
        'verify',
        'trust_env',
        # custom attrs
        '_base_url',
        '_mask_headers',
        '_mask_patterns',
        'log',
        'utils',
    ]

    def __init__(self, base_url: Optional[str] = None, logger: Optional[object] = None):
        """Initialize the Class properties."""
        super().__init__()
        self._base_url: str = base_url
        self.log: object = logger or logging.getLogger('session')

        self._custom_adapter: Optional[CustomAdapter] = None
        self.utils: object = Utils()

        # properties
        self._log_curl: bool = False
        self._mask_body = False
        self._mask_headers = True
        self._mask_patterns = None
        self._rate_limit_handler = RateLimitHandler()
        self._too_many_requests_handler = None

        # Add default Retry
        self.retry()

    @property
    def base_url(self) -> str:
        """Return the base url."""
        return self._base_url

    @base_url.setter
    def base_url(self, url):
        """Set base_url."""
        self._base_url = url.strip('/')

    @property
    def log_curl(self) -> bool:
        """Return whether or not requests will be logged as a curl command."""
        return self._log_curl

    @log_curl.setter
    def log_curl(self, log_curl: bool):
        """Enable or disable logging curl commands."""
        self._log_curl = log_curl

    @property
    def mask_body(self) -> bool:
        """Return property"""
        return self._mask_body

    @mask_body.setter
    def mask_body(self, mask_bool: bool):
        """Set property"""
        self._mask_body = mask_bool

    @property
    def mask_headers(self) -> bool:
        """Return property"""
        return self._mask_headers

    @mask_headers.setter
    def mask_headers(self, mask_bool: bool):
        """Set property"""
        self._mask_headers = mask_bool

    @property
    def mask_patterns(self) -> list:
        """Return property"""
        return self._mask_patterns

    @mask_patterns.setter
    def mask_patterns(self, patterns: list):
        """Set property"""
        self._mask_patterns = patterns

    @property
    def too_many_requests_handler(self) -> Callable[[Response], float]:
        """Get the too_many_requests_handler.

        The too_many_requests_handler is responsible for determining how long to sleep (in seconds)
        on a 429 response.  The default returns the value in the `Retry-After` header.
        """
        if not self._too_many_requests_handler:
            self._too_many_requests_handler = default_too_many_requests_handler
        return self._too_many_requests_handler

    @too_many_requests_handler.setter
    def too_many_requests_handler(self, too_many_requests_handler: Callable[[Response], float]):
        """Set the too_many_requests_handler.

        The too_many_requests_handler is responsible for determining how long to sleep (in seconds)
        on a 429 response.  The default returns the value in the `Retry-After` header.

        Args:
            too_many_requests_handler: callable that returns the number of seconds to wait on a
                429 response.
        """
        self._too_many_requests_handler = too_many_requests_handler

    @property
    def rate_limit_handler(self) -> RateLimitHandler:
        """Return the RateLimitHandler.

        The RateLimitHandler is responsible for throttling request frequency.  The default
        implementation uses X-RateLimit-Remaining and X-RateLimit-Reset headers.
        """
        return self._rate_limit_handler

    @rate_limit_handler.setter
    def rate_limit_handler(self, rate_limit_handler: RateLimitHandler):
        """Set the RateLimitHandler.

        The RateLimitHandler is responsible for throttling request frequency.  The default
        implementation uses X-RateLimit-Remaining and X-RateLimit-Reset headers.

        Args:
            rate_limit_handler: the RateLimitHandler object to use.
        """
        self._rate_limit_handler = rate_limit_handler
        if self._custom_adapter:
            self._custom_adapter.rate_limit_handler = rate_limit_handler

    def request(  # pylint: disable=arguments-differ
        self, method: str, url: str, **kwargs
    ) -> object:
        """Override request method disabling verify on token renewal if disabled on session.

        Args:
            method (str): The HTTP method
            url (str): The URL or path for the request.

        Returns:
            object: The requests Response object .
        """
        if self.base_url is not None and not url.startswith('https'):
            url = f'{self.base_url}{url}'

        # this kwargs value is used to signal 429 handling that this is a retry, but the super
        # method doesn't expect it so it needs to be removed.
        tc_is_retry = kwargs.pop('tc_is_retry', False)

        response: Response = super().request(method, url, **kwargs)

        if response.status_code == 429 and not tc_is_retry:
            too_many_requests_handler = self.too_many_requests_handler
            time.sleep(too_many_requests_handler(response))
            kwargs['tc_is_retry'] = True
            return self.request(method, url, **kwargs)

        # APP-79 - adding logging of request as curl commands
        if not response.ok or self.log_curl:
            try:
                self.log.debug(
                    self.utils.requests_to_curl(
                        response.request,
                        mask_body=self.mask_body,
                        mask_headers=self.mask_headers,
                        mask_patterns=self.mask_patterns,
                        proxies=self.proxies,
                        verify=self.verify,
                    )
                )
            except Exception:  # nosec
                pass  # logging curl command is best effort

        self.log.debug(
            f'feature=external-session, request-url={response.request.url}, '
            f'status_code={response.status_code}, elapsed={response.elapsed}'
        )

        return response

    def rate_limit_config(
        self,
        limit_remaining_header: str = 'X-RateLimit-Remaining',
        limit_reset_header: str = 'X-RateLimit-Reset',
        remaining_threshold: int = 0,
    ):
        """Configure rate-limiting.

        Configures the RateLimitHandler to use the given headers and remaining requests threshold.

        Args:
            limit_remaining_header: The header containing the number of requests remaining.
            limit_reset_header: The header that specifies when the rate limit period will reset.
            remaining_threshold: When the value in the limit_remaining_header is this value or
                lower, sleep until the time from the limit_reset_header.
        """
        self.rate_limit_handler.limit_remaining_header = limit_remaining_header
        self.rate_limit_handler.limit_reset_header = limit_reset_header
        self.rate_limit_handler.remaining_threshold = remaining_threshold

    def retry(
        self,
        retries: Optional[int] = 3,
        backoff_factor: Optional[float] = 0.3,
        status_forcelist: Optional[list] = None,
        **kwargs,
    ):
        """Add retry to Requests Session

        https://urllib3.readthedocs.io/en/latest/reference/urllib3.util.html#urllib3.util.retry.Retry

        Args:
            retries (Optional[int] = 3): The number of retry attempts.
            backoff_factor (Optional[float] = 0.3): The backoff factor for retries.
            status_forcelist (Optional[list] = [500, 502, 504]): A list of status code to retry on.
            urls (list, kwargs): An optional URL to apply the retry. If not provided the retry
                applies to all request with "https://".
        """
        retry_object: object = Retry(
            total=retries,
            read=retries,
            connect=retries,
            backoff_factor=backoff_factor,
            status_forcelist=status_forcelist or [500, 502, 504],
        )
        urls = kwargs.get('urls') or ['https://']

        if self._custom_adapter:
            self._custom_adapter.max_retries = retry_object
        else:
            self._custom_adapter = CustomAdapter(
                rate_limit_handler=self.rate_limit_handler, max_retries=retry_object
            )

        # mount the custom adapter
        for url in urls:
            self.log.info(
                f'feature=external-session, action=applying-retry, retries={retries}, '
                f'backoff-factor={backoff_factor}, status-forcelist={status_forcelist}, url={url}'
            )
            self.mount(url, self._custom_adapter)



"""TcEx Framework MQTT message broker module."""
# standard library
import ssl
import time
import traceback
from typing import List, Optional

# third-party
import paho.mqtt.client as mqtt


class MqttMessageBroker:
    """TcEx Framework MQTT message broker module."""

    def __init__(
        self,
        broker_host: str,
        broker_port: int,
        broker_timeout: int,
        logger: object,
        broker_token: Optional[str] = None,
        broker_cacert: Optional[str] = None,
    ):
        """Initialize the Class properties.

        Args:
            broker_host: The MQTT server host (IP).
            broker_port: The MQTT server port.
            broker_timeout: The MQTT connection timeout.
            broker_token: The MQTT connect token.
            broker_cacert: The CA certfile for connection.
            logger: A logging instance
        """
        self.broker_host = broker_host
        self.broker_port = int(broker_port)
        self.broker_timeout = int(broker_timeout)
        self.broker_token = broker_token
        self.broker_cacert = broker_cacert
        self.log = logger

        # properties
        self._client = None
        self._connected = False
        self._on_connect_callbacks: List[callable] = []
        self._on_disconnect_callbacks: List[callable] = []
        self._on_log_callbacks: List[callable] = []
        self._on_message_callbacks: List[callable] = []
        self._on_publish_callbacks: List[callable] = []
        self._on_subscribe_callbacks: List[callable] = []
        self._on_unsubscribe_callbacks: List[callable] = []
        self.shutdown = False  # used in service App for shutdown flag

    def add_on_connect_callback(self, callback: callable, index: Optional[int] = None) -> None:
        """Add a callback for on_connect events.

        Args:
            callback: A callback that matches signature of an on_connect event.
            index: The index value to insert the callback into the list.
        """
        index = index or len(self._on_connect_callbacks)
        self._on_connect_callbacks.insert(index, callback)

    def add_on_disconnect_callback(self, callback: callable, index: Optional[int] = None) -> None:
        """Add a callback for on_disconnect events.

        Args:
            callback: A callback that matches signature of an on_disconnect event.
            index: The index value to insert the callback into the list.
        """
        index = index or len(self._on_disconnect_callbacks)
        self._on_disconnect_callbacks.insert(callback)

    def add_on_log_callback(self, callback: callable, index: Optional[int] = None) -> None:
        """Add a callback for on_log events.

        Args:
            callback: A callback that matches signature of an on_log event.
            index: The index value to insert the callback into the list.
        """
        index = index or len(self._on_log_callbacks)
        self._on_log_callbacks.insert(callback)

    def add_on_message_callback(
        self, callback: callable, index: Optional[int] = None, topics: Optional[List[str]] = None
    ) -> None:
        """Add a callback for on_message events.

        Args:
            callback: A callback that matches signature of an on_message event.
            index: The index value to insert the callback into the list.
            topics: A optional list of topics to call callback. If value is None then callback
                will always be called.
        """
        index = index or len(self._on_message_callbacks)
        self._on_message_callbacks.insert(index, {'callback': callback, 'topics': topics})

    def add_on_publish_callback(self, callback: callable, index: Optional[int] = None) -> None:
        """Add a callback for on_publish events.

        Args:
            callback: A callback that matches signature of an on_publish event.
            index: The index value to insert the callback into the list.
        """
        index = index or len(self._on_publish_callbacks)
        self._on_publish_callbacks.insert(index, callback)

    def add_on_subscribe_callback(self, callback: callable, index: Optional[int] = None) -> None:
        """Add a callback for on_subscribe events.

        Args:
            callback: A callback that matches signature of an on_subscribe event.
            index: The index value to insert the callback into the list.
        """
        index = index or len(self._on_subscribe_callbacks)
        self._on_subscribe_callbacks.insert(index, callback)

    def add_on_unsubscribe_callback(self, callback: callable, index: Optional[int] = None) -> None:
        """Add a callback for on_unsubscribe events.

        Args:
            callback: A callback that matches signature of an on_unsubscribe event.
            index: The index value to insert the callback into the list.
        """
        index = index or len(self._on_unsubscribe_callbacks)
        self._on_unsubscribe_callbacks.insert(index, callback)

    @property
    def client(self) -> object:
        """Return MQTT client."""
        if self._client is None:
            try:
                self._client = mqtt.Client(client_id='', clean_session=True)
                self._client.connect(self.broker_host, self.broker_port, self.broker_timeout)
                if self.broker_cacert is not None:
                    self._client.tls_set(
                        ca_certs=self.broker_cacert,
                        cert_reqs=ssl.CERT_REQUIRED,
                        tls_version=ssl.PROTOCOL_TLSv1_2,
                    )
                    self._client.tls_insecure_set(False)
                # add logger
                # if self.log.getEffectiveLevel() == 5:
                #     self._client.enable_logger(logger=self.log)
                # username must be a empty string
                if self.broker_token is not None:
                    self._client.username_pw_set('', password=self.broker_token)
            except Exception as e:
                self.log.error(f'feature=message-broker, event=failed-connection, error="""{e}"""')
                self.log.trace(traceback.format_exc())
                self.shutdown = True

        return self._client

    def connect(self) -> None:
        """Listen for message coming from broker."""
        try:
            # handle connection issues by not using loop_forever. give the service X seconds to
            # connect to message broker, else timeout and log generic connection error.
            self.client.loop_start()
            deadline = time.time() + self.broker_timeout
            while True:
                if not self._connected and deadline < time.time():
                    self.client.loop_stop()
                    raise ConnectionError(
                        f'failed to connect to message broker host '
                        f'{self.broker_host} on port '
                        f'{self.broker_port}.'
                    )
                time.sleep(1)

        except Exception as e:
            self.log.trace(f'feature=message-broker, event=connection-error, error="""{e}"""')
            self.log.error(traceback.format_exc())

    def on_connect(self, client, userdata, flags, rc) -> None:
        """Handle MQTT on_connect events."""
        self.log.info(f'feature=message-broker, event=broker-connect, status={str(rc)}')
        self._connected = True
        for callback in self._on_connect_callbacks:
            callback(client, userdata, flags, rc)

    def on_disconnect(self, client, userdata, rc) -> None:
        """Handle MQTT on_disconnect events."""
        self.log.info(f'feature=message-broker, event=broker-disconnect, status={str(rc)}')
        for callback in self._on_disconnect_callbacks:
            callback(client, userdata, rc)

    def on_log(self, client, userdata, level, buf) -> None:
        """Handle MQTT on_log events."""
        # self.log.trace(f'feature=message-broker, event=on_log, buf={buf}, level={level}')
        for callback in self._on_log_callbacks:
            callback(client, userdata, level, buf)

    def on_message(self, client, userdata, message) -> None:
        """Handle MQTT on_message events."""
        mp = message.payload.decode().replace('\n', '')
        self.log.trace(
            f'''feature=message-broker, message-topic={message.topic}, message-payload={mp}'''
        )
        for cd in self._on_message_callbacks:
            topics = cd.get('topics')
            if topics is None or message.topic in topics:
                cd.get('callback')(client, userdata, message)

    def on_publish(self, client, userdata, result) -> None:
        """Handle MQTT on_publish events."""
        # self.log.trace(f'feature=message-broker, event=on_publish, result={result}')
        for callback in self._on_publish_callbacks:
            callback(client, userdata, result)

    def on_subscribe(self, client, userdata, mid, granted_qos) -> None:
        """Handle MQTT on_subscribe events."""
        # self.log.trace(
        #     f'feature=message-broker, event=on_subscribe, mid={mid}, granted_qos={granted_qos}'
        # )
        for callback in self._on_subscribe_callbacks:
            callback(client, userdata, mid, granted_qos)

    def on_unsubscribe(self, client, userdata, mid) -> None:
        """Handle MQTT on_unsubscribe events."""
        # self.log.trace(f'feature=message-broker, event=on_subscribe, mid={mid}')
        for callback in self._on_unsubscribe_callbacks:
            callback(client, userdata, mid)

    def publish(self, message: str, topic: str) -> None:
        """Publish a message on client topic.

        Args:
            message: The message to be sent on client topic.
            topic: The broker topic.
        """
        r: object = self.client.publish(topic, message)
        self.log.debug(
            f'feature=service, event=publish-message, topic="{topic}", '
            f'message={message}, response={r}'
        )

    def register_callbacks(self) -> None:
        """Register all the message broker callbacks."""
        self.client.on_connect = self.on_connect
        self.client.on_disconnect = self.on_disconnect
        self.client.on_log = self.on_log
        self.client.on_message = self.on_message
        self.client.on_publish = self.on_publish
        self.client.on_subscribe = self.on_subscribe
        self.client.on_unsubscribe = self.on_unsubscribe
